#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Contains definitions of standard gates such as
* Hadamard (H)
* Pauli-X (X / NOT)
* Pauli-Y (Y)
* Pauli-Z (Z)
* S and its inverse (S / Sdagger)
* T and its inverse (T / Tdagger)
* SqrtX gate (SqrtX)
* Swap gate (Swap)
* SqrtSwap gate (SqrtSwap)
* Entangle (Entangle)
* Phase gate (Ph)
* Rotation-X (Rx)
* Rotation-Y (Ry)
* Rotation-Z (Rz)
* Rotation-XX on two qubits (Rxx)
* Rotation-YY on two qubits (Ryy)
* Rotation-ZZ on two qubits (Rzz)
* Phase-shift (R)
* Measurement (Measure)

and meta gates, i.e.,
* Allocate / Deallocate qubits
* Flush gate (end of circuit)
* Barrier
* FlipBits
"""

import math
import cmath
import warnings

import numpy as np

from projectq.ops import get_inverse
from ._basics import (BasicGate, SelfInverseGate, BasicRotationGate,
                      BasicPhaseGate, ClassicalInstructionGate,
                      FastForwardingGate)
from ._command import apply_command


class HGate(SelfInverseGate):
    """ Hadamard gate class """

    def __str__(self):
        return "H"

    @property
    def matrix(self):
        return 1. / cmath.sqrt(2.) * np.matrix([[1, 1], [1, -1]])


#: Shortcut (instance of) :class:`projectq.ops.HGate`
H = HGate()


class XGate(SelfInverseGate):
    """ Pauli-X gate class """

    def __str__(self):
        return "X"

    @property
    def matrix(self):
        return np.matrix([[0, 1], [1, 0]])


#: Shortcut (instance of) :class:`projectq.ops.XGate`
X = NOT = XGate()


class YGate(SelfInverseGate):
    """ Pauli-Y gate class """

    def __str__(self):
        return "Y"

    @property
    def matrix(self):
        return np.matrix([[0, -1j], [1j, 0]])


#: Shortcut (instance of) :class:`projectq.ops.YGate`
Y = YGate()


class ZGate(SelfInverseGate):
    """ Pauli-Z gate class """

    def __str__(self):
        return "Z"

    @property
    def matrix(self):
        return np.matrix([[1, 0], [0, -1]])


#: Shortcut (instance of) :class:`projectq.ops.ZGate`
Z = ZGate()


class SGate(BasicGate):
    """ S gate class """

    @property
    def matrix(self):
        return np.matrix([[1, 0], [0, 1j]])

    def __str__(self):
        return "S"


#: Shortcut (instance of) :class:`projectq.ops.SGate`
S = SGate()
#: Inverse (and shortcut) of :class:`projectq.ops.SGate`
Sdag = Sdagger = get_inverse(S)


class TGate(BasicGate):
    """ T gate class """

    @property
    def matrix(self):
        return np.matrix([[1, 0], [0, cmath.exp(1j * cmath.pi / 4)]])

    def __str__(self):
        return "T"


#: Shortcut (instance of) :class:`projectq.ops.TGate`
T = TGate()
#: Inverse (and shortcut) of :class:`projectq.ops.TGate`
Tdag = Tdagger = get_inverse(T)


class SqrtXGate(BasicGate):
    """ Square-root X gate class """

    @property
    def matrix(self):
        return 0.5 * np.matrix([[1 + 1j, 1 - 1j], [1 - 1j, 1 + 1j]])

    def tex_str(self):
        return r'$\sqrt{X}$'

    def __str__(self):
        return "SqrtX"


#: Shortcut (instance of) :class:`projectq.ops.SqrtXGate`
SqrtX = SqrtXGate()


class SwapGate(SelfInverseGate):
    """ Swap gate class (swaps 2 qubits) """

    def __init__(self):
        SelfInverseGate.__init__(self)
        self.interchangeable_qubit_indices = [[0, 1]]

    def __str__(self):
        return "Swap"

    @property
    def matrix(self):
        return np.matrix([[1, 0, 0, 0],
                          [0, 0, 1, 0],
                          [0, 1, 0, 0],
                          [0, 0, 0, 1]])


#: Shortcut (instance of) :class:`projectq.ops.SwapGate`
Swap = SwapGate()


class SqrtSwapGate(BasicGate):
    """ Square-root Swap gate class """

    def __init__(self):
        BasicGate.__init__(self)
        self.interchangeable_qubit_indices = [[0, 1]]

    def __str__(self):
        return "SqrtSwap"

    @property
    def matrix(self):
        return np.matrix([[1, 0, 0, 0],
                          [0, 0.5+0.5j, 0.5-0.5j, 0],
                          [0, 0.5-0.5j, 0.5+0.5j, 0],
                          [0, 0, 0, 1]])


#: Shortcut (instance of) :class:`projectq.ops.SqrtSwapGate`
SqrtSwap = SqrtSwapGate()


class EntangleGate(BasicGate):
    """
    Entangle gate (Hadamard on first qubit, followed by CNOTs applied to all
    other qubits).
    """

    def __str__(self):
        return "Entangle"


#: Shortcut (instance of) :class:`projectq.ops.EntangleGate`
Entangle = EntangleGate()


class Ph(BasicPhaseGate):
    """ Phase gate (global phase) """

    @property
    def matrix(self):
        return np.matrix([[cmath.exp(1j * self.angle), 0],
                          [0, cmath.exp(1j * self.angle)]])


class Rx(BasicRotationGate):
    """ RotationX gate class """

    @property
    def matrix(self):
        return np.matrix([[math.cos(0.5 * self.angle),
                           -1j * math.sin(0.5 * self.angle)],
                          [-1j * math.sin(0.5 * self.angle),
                           math.cos(0.5 * self.angle)]])


class Ry(BasicRotationGate):
    """ RotationY gate class """
    @property
    def matrix(self):
        return np.matrix([[math.cos(0.5 * self.angle),
                           -math.sin(0.5 * self.angle)],
                          [math.sin(0.5 * self.angle),
                           math.cos(0.5 * self.angle)]])


class Rz(BasicRotationGate):
    """ RotationZ gate class """

    @property
    def matrix(self):
        return np.matrix([[cmath.exp(-.5 * 1j * self.angle), 0],
                          [0, cmath.exp(.5 * 1j * self.angle)]])


class Rxx(BasicRotationGate):
    """ RotationXX gate class """
    @property
    def matrix(self):
        return np.matrix([
            [cmath.cos(.5 * self.angle), 0, 0, -1j*cmath.sin(.5 * self.angle)],
            [0, cmath.cos(.5 * self.angle), -1j*cmath.sin(.5 * self.angle), 0],
            [0, -1j*cmath.sin(.5 * self.angle), cmath.cos(.5 * self.angle), 0],
            [-1j*cmath.sin(.5 * self.angle), 0, 0, cmath.cos(.5 * self.angle)]
        ])


class Ryy(BasicRotationGate):
    """ RotationYY gate class """
    @property
    def matrix(self):
        return np.matrix([
            [cmath.cos(.5*self.angle), 0, 0, 1j*cmath.sin(.5 * self.angle)],
            [0, cmath.cos(.5 * self.angle), -1j*cmath.sin(.5 * self.angle), 0],
            [0, -1j*cmath.sin(.5 * self.angle), cmath.cos(.5 * self.angle), 0],
            [1j*cmath.sin(.5 * self.angle), 0, 0, cmath.cos(.5 * self.angle)]])


class Rzz(BasicRotationGate):
    """ RotationZZ gate class """
    @property
    def matrix(self):
        return np.matrix([
            [cmath.exp(-.5 * 1j * self.angle), 0, 0, 0],
            [0, cmath.exp(.5 * 1j * self.angle), 0, 0],
            [0, 0, cmath.exp(.5 * 1j * self.angle), 0],
            [0, 0, 0, cmath.exp(-.5 * 1j * self.angle)]])


class R(BasicPhaseGate):
    """ Phase-shift gate (equivalent to Rz up to a global phase) """

    @property
    def matrix(self):
        return np.matrix([[1, 0], [0, cmath.exp(1j * self.angle)]])


class FlushGate(FastForwardingGate):
    """
    Flush gate (denotes the end of the circuit).

    Note:
        All compiler engines (cengines) which cache/buffer gates are obligated
        to flush and send all gates to the next compiler engine (followed by
        the flush command).

    Note:
        This gate is sent when calling

        .. code-block:: python

            eng.flush()

        on the MainEngine `eng`.
    """

    def __str__(self):
        return ""


class MeasureGate(FastForwardingGate):
    """ Measurement gate class (for single qubits)."""

    def __str__(self):
        return "Measure"

    def __or__(self, qubits):
        """
        Previously (ProjectQ <= v0.3.6) MeasureGate/Measure was allowed to be
        applied to any number of quantum registers. Now the MeasureGate/Measure
        is strictly a single qubit gate. In the coming releases the backward
        compatibility will be removed!
        """
        num_qubits = 0
        for qureg in self.make_tuple_of_qureg(qubits):
            for qubit in qureg:
                num_qubits += 1
                cmd = self.generate_command(([qubit], ))
                apply_command(cmd)
        if num_qubits > 1:
            warnings.warn("Pending syntax change in future versions of "
                          "ProjectQ: \n Measure will be a single qubit gate "
                          "only. Use `All(Measure) | qureg` instead to "
                          "measure multiple qubits.")


#: Shortcut (instance of) :class:`projectq.ops.MeasureGate`
Measure = MeasureGate()


class AllocateQubitGate(ClassicalInstructionGate):
    """ Qubit allocation gate class """

    def __str__(self):
        return "Allocate"

    def get_inverse(self):
        return DeallocateQubitGate()


#: Shortcut (instance of) :class:`projectq.ops.AllocateQubitGate`
Allocate = AllocateQubitGate()


class DeallocateQubitGate(FastForwardingGate):
    """ Qubit deallocation gate class """

    def __str__(self):
        return "Deallocate"

    def get_inverse(self):
        return Allocate


#: Shortcut (instance of) :class:`projectq.ops.DeallocateQubitGate`
Deallocate = DeallocateQubitGate()


class AllocateDirtyQubitGate(ClassicalInstructionGate):
    """ Dirty qubit allocation gate class """

    def __str__(self):
        return "AllocateDirty"

    def get_inverse(self):
        return Deallocate


#: Shortcut (instance of) :class:`projectq.ops.AllocateDirtyQubitGate`
AllocateDirty = AllocateDirtyQubitGate()


class BarrierGate(BasicGate):
    """ Barrier gate class """

    def __str__(self):
        return "Barrier"

    def get_inverse(self):
        return Barrier


#: Shortcut (instance of) :class:`projectq.ops.BarrierGate`
Barrier = BarrierGate()


class FlipBits(SelfInverseGate):
    """ Gate for flipping qubits by means of XGates """
    def __init__(self, bits_to_flip):
        """
        Initialize FlipBits gate.

        Example:
            .. code-block:: python

                qureg = eng.allocate_qureg(2)
                FlipBits([0, 1]) | qureg

        Args:
            bits_to_flip(list[int]|list[bool]|str|int): int or array of 0/1,
               True/False, or string of 0/1 identifying the qubits to flip.
               In case of int, the bits to flip are determined from the
               binary digits, with the least significant bit corresponding
               to qureg[0]. If bits_to_flip is negative, exactly all qubits
               which would not be flipped for the input -bits_to_flip-1 are
               flipped, i.e., bits_to_flip=-1 flips all qubits.
        """
        SelfInverseGate.__init__(self)
        if isinstance(bits_to_flip, int):
            self.bits_to_flip = bits_to_flip
        else:
            self.bits_to_flip = 0
            for i in reversed(list(bits_to_flip)):
                bit = 0b1 if i == '1' or i == 1 or i is True else 0b0
                self.bits_to_flip = (self.bits_to_flip << 1) | bit

    def __str__(self):
        return "FlipBits("+str(self.bits_to_flip)+")"

    def __or__(self, qubits):
        quregs_tuple = self.make_tuple_of_qureg(qubits)
        if len(quregs_tuple) > 1:
            raise ValueError(self.__str__()+' can only be applied to qubits,'
                             'quregs, arrays of qubits, and tuples with one'
                             'individual qubit')
        for qureg in quregs_tuple:
            for i, qubit in enumerate(qureg):
                if (self.bits_to_flip >> i) & 1:
                    XGate() | qubit

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.bits_to_flip == other.bits_to_flip
        return False

    def __hash__(self):
        return hash(self.__str__())



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


def _exec(code, qs):
    """
    Executes the Python code in 'filename'.

    Args:
        code (string): ProjectQ code.
        qs (tuple<Qureg>): Qubits to which the permutation is being
                           applied.
    """
    from projectq.ops import C, X, Z, All
    exec(code)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Contains a compiler engine which generates TikZ Latex code describing the
circuit.
"""
from builtins import input

from projectq.cengines import LastEngineException, BasicEngine
from projectq.ops import FlushGate, Measure, Allocate, Deallocate
from projectq.meta import get_control_count
from projectq.backends._circuits import to_latex


class CircuitItem(object):
    def __init__(self, gate, lines, ctrl_lines):
        """
        Initialize a circuit item.

        Args:
            gate: Gate object.
            lines (list<int>): Circuit lines the gate acts on.
            ctrl_lines (list<int>): Circuit lines which control the gate.
        """
        self.gate = gate
        self.lines = lines
        self.ctrl_lines = ctrl_lines
        self.id = -1

    def __eq__(self, other):
        return (self.gate == other.gate and self.lines == other.lines
                and self.ctrl_lines == other.ctrl_lines
                and self.id == other.id)

    def __ne__(self, other):
        return not self.__eq__(other)


class CircuitDrawer(BasicEngine):
    """
    CircuitDrawer is a compiler engine which generates TikZ code for drawing
    quantum circuits.

    The circuit can be modified by editing the settings.json file which is
    generated upon first execution. This includes adjusting the gate width,
    height, shadowing, line thickness, and many more options.

    After initializing the CircuitDrawer, it can also be given the mapping
    from qubit IDs to wire location (via the :meth:`set_qubit_locations`
    function):

    .. code-block:: python

        circuit_backend = CircuitDrawer()
        circuit_backend.set_qubit_locations({0: 1, 1: 0}) # swap lines 0 and 1
        eng = MainEngine(circuit_backend)

        ... # run quantum algorithm on this main engine

        print(circuit_backend.get_latex()) # prints LaTeX code

    To see the qubit IDs in the generated circuit, simply set the `draw_id`
    option in the settings.json file under "gates":"AllocateQubitGate" to
    True:

    .. code-block:: python

        "gates": {
            "AllocateQubitGate": {
                "draw_id": True,
                "height": 0.15,
                "width": 0.2,
                "pre_offset": 0.1,
                "offset": 0.1
            },
            ...

    The settings.json file has the following structure:

    .. code-block:: python

        {
            "control": { # settings for control "circle"
                    "shadow": false,
                    "size": 0.1
            },
            "gate_shadow": true, # enable/disable shadows for all gates
            "gates": {
                    "GateClassString": {
                        GATE_PROPERTIES
                    }
                    "GateClassString2": {
                        ...
            },
            "lines": { # settings for qubit lines
                    "double_classical": true, # draw double-lines for
                                              # classical bits
                    "double_lines_sep": 0.04, # gap between the two lines
                                              # for double lines
                    "init_quantum": true, # start out with quantum bits
                    "style": "very thin" # line style
            }
        }

    All gates (except for the ones requiring special treatment) support the
    following properties:

    .. code-block:: python

        "GateClassString": {
            "height": GATE_HEIGHT,
            "width": GATE_WIDTH
            "pre_offset": OFFSET_BEFORE_PLACEMENT,
            "offset": OFFSET_AFTER_PLACEMENT,
        },

    """
    def __init__(self, accept_input=False, default_measure=0):
        """
        Initialize a circuit drawing engine.

        The TikZ code generator uses a settings file (settings.json), which
        can be altered by the user. It contains gate widths, heights, offsets,
        etc.

        Args:
            accept_input (bool): If accept_input is true, the printer queries
                the user to input measurement results if the CircuitDrawer is
                the last engine. Otherwise, all measurements yield the result
                default_measure (0 or 1).
            default_measure (bool): Default value to use as measurement
                results if accept_input is False and there is no underlying
                backend to register real measurement results.
        """
        BasicEngine.__init__(self)
        self._accept_input = accept_input
        self._default_measure = default_measure
        self._qubit_lines = dict()
        self._free_lines = []
        self._map = dict()

        # Order in which qubit lines are drawn
        self._drawing_order = []

    def is_available(self, cmd):
        """
        Specialized implementation of is_available: Returns True if the
        CircuitDrawer is the last engine (since it can print any command).

        Args:
            cmd (Command): Command for which to check availability (all
                Commands can be printed).
        Returns:
            availability (bool): True, unless the next engine cannot handle
            the Command (if there is a next engine).
        """
        try:
            return BasicEngine.is_available(self, cmd)
        except LastEngineException:
            return True

    def set_qubit_locations(self, id_to_loc):
        """
        Sets the qubit lines to use for the qubits explicitly.

        To figure out the qubit IDs, simply use the setting `draw_id` in the
        settings file. It is located in "gates":"AllocateQubitGate".
        If draw_id is True, the qubit IDs are drawn in red.

        Args:
            id_to_loc (dict): Dictionary mapping qubit ids to qubit line
                numbers.

        Raises:
            RuntimeError: If the mapping has already begun (this function
                needs be called before any gates have been received).
        """
        if len(self._map) > 0:
            raise RuntimeError("set_qubit_locations() has to be called before"
                               " applying gates!")

        for k in range(min(id_to_loc), max(id_to_loc) + 1):
            if k not in id_to_loc:
                raise RuntimeError("set_qubit_locations(): Invalid id_to_loc "
                                   "mapping provided. All ids in the provided"
                                   " range of qubit ids have to be mapped "
                                   "somewhere.")
        self._map = id_to_loc

    def _print_cmd(self, cmd):
        """
        Add the command cmd to the circuit diagram, taking care of potential
        measurements as specified in the __init__ function.

        Queries the user for measurement input if a measurement command
        arrives if accept_input was set to True. Otherwise, it uses the
        default_measure parameter to register the measurement outcome.

        Args:
            cmd (Command): Command to add to the circuit diagram.
        """
        if cmd.gate == Allocate:
            qubit_id = cmd.qubits[0][0].id
            if qubit_id not in self._map:
                self._map[qubit_id] = qubit_id
            self._qubit_lines[qubit_id] = []

        if cmd.gate == Deallocate:
            qubit_id = cmd.qubits[0][0].id
            self._free_lines.append(qubit_id)

        if self.is_last_engine and cmd.gate == Measure:
            assert get_control_count(cmd) == 0

            for qureg in cmd.qubits:
                for qubit in qureg:
                    if self._accept_input:
                        m = None
                        while m not in ('0', '1', 1, 0):
                            prompt = ("Input measurement result (0 or 1) for "
                                      "qubit " + str(qubit) + ": ")
                            m = input(prompt)
                    else:
                        m = self._default_measure
                    m = int(m)
                    self.main_engine.set_measurement_result(qubit, m)

        all_lines = [qb.id for qr in cmd.all_qubits for qb in qr]

        gate = cmd.gate
        lines = [qb.id for qr in cmd.qubits for qb in qr]
        ctrl_lines = [qb.id for qb in cmd.control_qubits]
        item = CircuitItem(gate, lines, ctrl_lines)
        for l in all_lines:
            self._qubit_lines[l].append(item)

        self._drawing_order.append(all_lines[0])

    def get_latex(self, ordered=False, draw_gates_in_parallel=True):
        """
        Return the latex document string representing the circuit.

        Simply write this string into a tex-file or, alternatively, pipe the
        output directly to, e.g., pdflatex:

        .. code-block:: bash

            python3 my_circuit.py | pdflatex

        where my_circuit.py calls this function and prints it to the terminal.

        Args:
            ordered(bool): flag if the gates should be drawn in the order they
                were added to the circuit
            draw_gates_in_parallel(bool): flag if parallel gates should be
                drawn parallel (True), or not (False)
        """
        qubit_lines = dict()

        for line in range(len(self._qubit_lines)):
            new_line = self._map[line]
            qubit_lines[new_line] = []
            for cmd in self._qubit_lines[line]:
                lines = [self._map[qb_id] for qb_id in cmd.lines]
                ctrl_lines = [self._map[qb_id] for qb_id in cmd.ctrl_lines]
                gate = cmd.gate
                new_cmd = CircuitItem(gate, lines, ctrl_lines)
                if gate == Allocate:
                    new_cmd.id = cmd.lines[0]
                qubit_lines[new_line].append(new_cmd)

        drawing_order = None
        if ordered:
            drawing_order = self._drawing_order

        return to_latex(qubit_lines,
                        drawing_order=drawing_order,
                        draw_gates_in_parallel=draw_gates_in_parallel)

    def receive(self, command_list):
        """
        Receive a list of commands from the previous engine, print the
        commands, and then send them on to the next engine.

        Args:
            command_list (list<Command>): List of Commands to print (and
                potentially send on to the next engine).
        """
        for cmd in command_list:
            if not cmd.gate == FlushGate():
                self._print_cmd(cmd)
            # (try to) send on
            if not self.is_last_engine:
                self.send([cmd])



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
A simulator that only permits classical operations, for faster/easier testing.
"""

from projectq.cengines import BasicEngine
from projectq.meta import LogicalQubitIDTag
from projectq.ops import (XGate,
                          BasicMathGate,
                          Measure,
                          FlushGate,
                          Allocate,
                          Deallocate)
from projectq.types import WeakQubitRef


class ClassicalSimulator(BasicEngine):
    """
    A simple introspective simulator that only permits classical operations.

    Allows allocation, deallocation, measuring (no-op), flushing (no-op),
    controls, NOTs, and any BasicMathGate. Supports reading/writing directly
    from/to bits and registers of bits.
    """
    def __init__(self):
        BasicEngine.__init__(self)
        self._state = 0
        self._bit_positions = {}

    def _convert_logical_to_mapped_qubit(self, qubit):
        """
        Converts a qubit from a logical to a mapped qubit if there is a mapper.

        Args:
            qubit (projectq.types.Qubit): Logical quantum bit
        """
        mapper = self.main_engine.mapper
        if mapper is not None:
            if qubit.id not in mapper.current_mapping:
                raise RuntimeError("Unknown qubit id. "
                                   "Please make sure you have called "
                                   "eng.flush().")
            return WeakQubitRef(qubit.engine,
                                mapper.current_mapping[qubit.id])
        return qubit

    def read_bit(self, qubit):
        """
        Reads a bit.

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.

        Args:
            qubit (projectq.types.Qubit): The bit to read.

        Returns:
            int: 0 if the target bit is off, 1 if it's on.
        """
        qubit = self._convert_logical_to_mapped_qubit(qubit)
        return self._read_mapped_bit(qubit)

    def _read_mapped_bit(self, mapped_qubit):
        """ Internal use only. Does not change logical to mapped qubits."""
        p = self._bit_positions[mapped_qubit.id]
        return (self._state >> p) & 1

    def write_bit(self, qubit, value):
        """
        Resets/sets a bit to the given value.

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.

        Args:
            qubit (projectq.types.Qubit): The bit to write.
            value (bool|int): Writes 1 if this value is truthy, else 0.
        """
        qubit = self._convert_logical_to_mapped_qubit(qubit)
        self._write_mapped_bit(qubit, value)

    def _write_mapped_bit(self, mapped_qubit, value):
        """ Internal use only. Does not change logical to mapped qubits."""
        p = self._bit_positions[mapped_qubit.id]
        if value:
            self._state |= 1 << p
        else:
            self._state &= ~(1 << p)

    def _mask(self, qureg):
        """
        Returns a mask, to compare against the state, with bits from the
        register set to 1 and other bits set to 0.

        Args:
            qureg (projectq.types.Qureg):
                The bits whose positions should be set.

        Returns:
            int: The mask.
        """
        t = 0
        for q in qureg:
            t |= 1 << self._bit_positions[q.id]
        return t

    def read_register(self, qureg):
        """
        Reads a group of bits as a little-endian integer.

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.

        Args:
            qureg (projectq.types.Qureg):
                The group of bits to read, in little-endian order.

        Returns:
            int: Little-endian register value.
        """
        new_qureg = []
        for qubit in qureg:
            new_qureg.append(self._convert_logical_to_mapped_qubit(qubit))
        return self._read_mapped_register(new_qureg)

    def _read_mapped_register(self, mapped_qureg):
        """ Internal use only. Does not change logical to mapped qubits."""
        t = 0
        for i in range(len(mapped_qureg)):
            t |= self._read_mapped_bit(mapped_qureg[i]) << i
        return t

    def write_register(self, qureg, value):
        """
        Sets a group of bits to store a little-endian integer value.

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.

        Args:
            qureg (projectq.types.Qureg):
                The bits to write, in little-endian order.
            value (int): The integer value to store. Must fit in the register.
        """
        new_qureg = []
        for qubit in qureg:
            new_qureg.append(self._convert_logical_to_mapped_qubit(qubit))
        self._write_mapped_register(new_qureg, value)

    def _write_mapped_register(self, mapped_qureg, value):
        """ Internal use only. Does not change logical to mapped qubits."""
        if value < 0 or value >= 1 << len(mapped_qureg):
            raise ValueError("Value won't fit in register.")
        for i in range(len(mapped_qureg)):
            self._write_mapped_bit(mapped_qureg[i], (value >> i) & 1)

    def is_available(self, cmd):
        return (cmd.gate == Measure or
                cmd.gate == Allocate or
                cmd.gate == Deallocate or
                isinstance(cmd.gate, (BasicMathGate, FlushGate, XGate)))

    def receive(self, command_list):
        for cmd in command_list:
            self._handle(cmd)
        if not self.is_last_engine:
            self.send(command_list)

    def _handle(self, cmd):
        if isinstance(cmd.gate, FlushGate):
            return

        if cmd.gate == Measure:
            for qr in cmd.qubits:
                for qb in qr:
                    # Check if a mapper assigned a different logical id
                    logical_id_tag = None
                    for tag in cmd.tags:
                        if isinstance(tag, LogicalQubitIDTag):
                            logical_id_tag = tag
                    log_qb = qb
                    if logical_id_tag is not None:
                        log_qb = WeakQubitRef(qb.engine,
                                              logical_id_tag.logical_qubit_id)
                    self.main_engine.set_measurement_result(
                        log_qb, self._read_mapped_bit(qb))
            return

        if cmd.gate == Allocate:
            new_id = cmd.qubits[0][0].id
            self._bit_positions[new_id] = len(self._bit_positions)
            return

        if cmd.gate == Deallocate:
            old_id = cmd.qubits[0][0].id
            pos = self._bit_positions[old_id]
            low = (1 << pos) - 1
            self._state = (self._state & low) | ((self._state >> 1) & ~low)
            self._bit_positions = {
                k: b - (0 if b < pos else 1)
                for k, b in self._bit_positions.items()
            }
            return

        controls_mask = self._mask(cmd.control_qubits)
        meets_controls = self._state & controls_mask == controls_mask

        if isinstance(cmd.gate, XGate):
            assert len(cmd.qubits) == 1 and len(cmd.qubits[0]) == 1
            target = cmd.qubits[0][0]
            if meets_controls:
                self._write_mapped_bit(target,
                                       not self._read_mapped_bit(target))
            return

        if isinstance(cmd.gate, BasicMathGate):
            if meets_controls:
                ins = [self._read_mapped_register(reg) for reg in cmd.qubits]
                outs = cmd.gate.get_math_function(cmd.qubits)(ins)
                for reg, out in zip(cmd.qubits, outs):
                    self._write_mapped_register(reg,
                                                out & ((1 << len(reg)) - 1))
            return

        raise ValueError("Only support alloc/dealloc/measure/not/math ops.")



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Defines a setup useful for the IBM QE chip with 16 qubits.

It provides the `engine_list` for the `MainEngine`, and contains an
AutoReplacer with most of the gate decompositions of ProjectQ, among others
it includes:

    * Controlled z-rotations --> Controlled NOTs and single-qubit rotations
    * Toffoli gate --> CNOT and single-qubit gates
    * m-Controlled global phases --> (m-1)-controlled phase-shifts
    * Global phases --> ignore
    * (controlled) Swap gates --> CNOTs and Toffolis
    * Arbitrary single qubit gates --> Rz and Ry
    * Controlled arbitrary single qubit gates --> Rz, Ry, and CNOT gates

Moreover, it contains `LocalOptimizers`.
"""

import projectq
import projectq.libs.math
import projectq.setups.decompositions
from projectq.cengines import (AutoReplacer,
                               DecompositionRuleSet,
                               GridMapper,
                               InstructionFilter,
                               LocalOptimizer,
                               SwapAndCNOTFlipper,
                               TagRemover)
from projectq.setups.grid import high_level_gates


ibmqx5_connections = set([(1, 0), (1, 2), (2, 3), (3, 4), (3, 14), (5, 4),
                          (6, 5), (6, 7), (6, 11), (7, 10), (8, 7), (9, 8),
                          (9, 10), (11, 10), (12, 5), (12, 11), (12, 13),
                          (13, 4), (13, 14), (15, 0), (15, 2), (15, 14)])


grid_to_physical = {0: 1, 1: 2, 2: 3, 3: 4, 4: 5, 5: 6, 6: 7, 7: 8, 8: 0,
                    9: 15, 10: 14, 11: 13, 12: 12, 13: 11, 14: 10, 15: 9}


def get_engine_list():
    rule_set = DecompositionRuleSet(modules=[projectq.libs.math,
                                             projectq.setups.decompositions])
    return [TagRemover(),
            LocalOptimizer(5),
            AutoReplacer(rule_set),
            InstructionFilter(high_level_gates),
            TagRemover(),
            LocalOptimizer(5),
            AutoReplacer(rule_set),
            TagRemover(),
            GridMapper(2, 8, grid_to_physical),
            LocalOptimizer(5),
            SwapAndCNOTFlipper(ibmqx5_connections),
            LocalOptimizer(5)]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains the projectq interface to a C++-based simulator, which has to be
built first. If the c++ simulator is not exported to python, a (slow) python
implementation is used as an alternative.
"""

import math
import random
from projectq.cengines import BasicEngine
from projectq.meta import get_control_count, LogicalQubitIDTag
from projectq.ops import (Measure, FlushGate, Allocate, Deallocate,
                          BasicMathGate, TimeEvolution)
from projectq.types import WeakQubitRef

FALLBACK_TO_PYSIM = False
try:
    from ._cppsim import Simulator as SimulatorBackend
except ImportError:
    from ._pysim import Simulator as SimulatorBackend
    FALLBACK_TO_PYSIM = True


class Simulator(BasicEngine):
    """
    Simulator is a compiler engine which simulates a quantum computer using
    C++-based kernels.

    OpenMP is enabled and the number of threads can be controlled using the
    OMP_NUM_THREADS environment variable, i.e.

    .. code-block:: bash

        export OMP_NUM_THREADS=4 # use 4 threads
        export OMP_PROC_BIND=spread # bind threads to processors by spreading
    """
    def __init__(self, gate_fusion=False, rnd_seed=None):
        """
        Construct the C++/Python-simulator object and initialize it with a
        random seed.

        Args:
            gate_fusion (bool): If True, gates are cached and only executed
                once a certain gate-size has been reached (only has an effect
                for the c++ simulator).
            rnd_seed (int): Random seed (uses random.randint(0, 4294967295) by
                default).

        Example of gate_fusion: Instead of applying a Hadamard gate to 5
        qubits, the simulator calculates the kronecker product of the 1-qubit
        gate matrices and then applies one 5-qubit gate. This increases
        operational intensity and keeps the simulator from having to iterate
        through the state vector multiple times. Depending on the system (and,
        especially, number of threads), this may or may not be beneficial.

        Note:
            If the C++ Simulator extension was not built or cannot be found,
            the Simulator defaults to a Python implementation of the kernels.
            While this is much slower, it is still good enough to run basic
            quantum algorithms.

            If you need to run large simulations, check out the tutorial in
            the docs which gives futher hints on how to build the C++
            extension.
        """
        if rnd_seed is None:
            rnd_seed = random.randint(0, 4294967295)
        BasicEngine.__init__(self)
        self._simulator = SimulatorBackend(rnd_seed)
        self._gate_fusion = gate_fusion

    def is_available(self, cmd):
        """
        Specialized implementation of is_available: The simulator can deal
        with all arbitrarily-controlled gates which provide a
        gate-matrix (via gate.matrix) and acts on 5 or less qubits (not
        counting the control qubits).

        Args:
            cmd (Command): Command for which to check availability (single-
                qubit gate, arbitrary controls)

        Returns:
            True if it can be simulated and False otherwise.
        """
        if (cmd.gate == Measure or cmd.gate == Allocate or
                cmd.gate == Deallocate or
                isinstance(cmd.gate, (BasicMathGate, TimeEvolution))):
            return True
        try:
            m = cmd.gate.matrix
            # Allow up to 5-qubit gates
            if len(m) > 2 ** 5:
                return False
            return True
        except AttributeError:
            return False

    def _convert_logical_to_mapped_qureg(self, qureg):
        """
        Converts a qureg from logical to mapped qubits if there is a mapper.

        Args:
            qureg (list[Qubit],Qureg): Logical quantum bits
        """
        mapper = self.main_engine.mapper
        if mapper is not None:
            mapped_qureg = []
            for qubit in qureg:
                if qubit.id not in mapper.current_mapping:
                    raise RuntimeError("Unknown qubit id. "
                                       "Please make sure you have called "
                                       "eng.flush().")
                new_qubit = WeakQubitRef(qubit.engine,
                                         mapper.current_mapping[qubit.id])
                mapped_qureg.append(new_qubit)
            return mapped_qureg
        return qureg

    def get_expectation_value(self, qubit_operator, qureg):
        """
        Get the expectation value of qubit_operator w.r.t. the current wave
        function represented by the supplied quantum register.

        Args:
            qubit_operator (projectq.ops.QubitOperator): Operator to measure.
            qureg (list[Qubit],Qureg): Quantum bits to measure.

        Returns:
            Expectation value

        Note:
            Make sure all previous commands (especially allocations) have
            passed through the compilation chain (call main_engine.flush() to
            make sure).

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.

        Raises:
            Exception: If `qubit_operator` acts on more qubits than present in
                the `qureg` argument.
        """
        qureg = self._convert_logical_to_mapped_qureg(qureg)
        num_qubits = len(qureg)
        for term, _ in qubit_operator.terms.items():
            if not term == () and term[-1][0] >= num_qubits:
                raise Exception("qubit_operator acts on more qubits than "
                                "contained in the qureg.")
        operator = [(list(term), coeff) for (term, coeff)
                    in qubit_operator.terms.items()]
        return self._simulator.get_expectation_value(operator,
                                                     [qb.id for qb in qureg])

    def apply_qubit_operator(self, qubit_operator, qureg):
        """
        Apply a (possibly non-unitary) qubit_operator to the current wave
        function represented by the supplied quantum register.

        Args:
            qubit_operator (projectq.ops.QubitOperator): Operator to apply.
            qureg (list[Qubit],Qureg): Quantum bits to which to apply the
                operator.

        Raises:
            Exception: If `qubit_operator` acts on more qubits than present in
                the `qureg` argument.

        Warning:
            This function allows applying non-unitary gates and it will not
            re-normalize the wave function! It is for numerical experiments
            only and should not be used for other purposes.

        Note:
            Make sure all previous commands (especially allocations) have
            passed through the compilation chain (call main_engine.flush() to
            make sure).

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.
        """
        qureg = self._convert_logical_to_mapped_qureg(qureg)
        num_qubits = len(qureg)
        for term, _ in qubit_operator.terms.items():
            if not term == () and term[-1][0] >= num_qubits:
                raise Exception("qubit_operator acts on more qubits than "
                                "contained in the qureg.")
        operator = [(list(term), coeff) for (term, coeff)
                    in qubit_operator.terms.items()]
        return self._simulator.apply_qubit_operator(operator,
                                                    [qb.id for qb in qureg])

    def get_probability(self, bit_string, qureg):
        """
        Return the probability of the outcome `bit_string` when measuring
        the quantum register `qureg`.

        Args:
            bit_string (list[bool|int]|string[0|1]): Measurement outcome.
            qureg (Qureg|list[Qubit]): Quantum register.

        Returns:
            Probability of measuring the provided bit string.

        Note:
            Make sure all previous commands (especially allocations) have
            passed through the compilation chain (call main_engine.flush() to
            make sure).

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.
        """
        qureg = self._convert_logical_to_mapped_qureg(qureg)
        bit_string = [bool(int(b)) for b in bit_string]
        return self._simulator.get_probability(bit_string,
                                               [qb.id for qb in qureg])

    def get_amplitude(self, bit_string, qureg):
        """
        Return the probability amplitude of the supplied `bit_string`.
        The ordering is given by the quantum register `qureg`, which must
        contain all allocated qubits.

        Args:
            bit_string (list[bool|int]|string[0|1]): Computational basis state
            qureg (Qureg|list[Qubit]): Quantum register determining the
                ordering. Must contain all allocated qubits.

        Returns:
            Probability amplitude of the provided bit string.

        Note:
            Make sure all previous commands (especially allocations) have
            passed through the compilation chain (call main_engine.flush() to
            make sure).

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.
        """
        qureg = self._convert_logical_to_mapped_qureg(qureg)
        bit_string = [bool(int(b)) for b in bit_string]
        return self._simulator.get_amplitude(bit_string,
                                             [qb.id for qb in qureg])

    def set_wavefunction(self, wavefunction, qureg):
        """
        Set the wavefunction and the qubit ordering of the simulator.

        The simulator will adopt the ordering of qureg (instead of reordering
        the wavefunction).

        Args:
            wavefunction (list[complex]): Array of complex amplitudes
                describing the wavefunction (must be normalized).
            qureg (Qureg|list[Qubit]): Quantum register determining the
                ordering. Must contain all allocated qubits.

        Note:
            Make sure all previous commands (especially allocations) have
            passed through the compilation chain (call main_engine.flush() to
            make sure).

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.
        """
        qureg = self._convert_logical_to_mapped_qureg(qureg)
        self._simulator.set_wavefunction(wavefunction,
                                         [qb.id for qb in qureg])

    def collapse_wavefunction(self, qureg, values):
        """
        Collapse a quantum register onto a classical basis state.

        Args:
            qureg (Qureg|list[Qubit]): Qubits to collapse.
            values (list[bool|int]|string[0|1]): Measurement outcome for each
                                                 of the qubits in `qureg`.

        Raises:
            RuntimeError: If an outcome has probability (approximately) 0 or
                if unknown qubits are provided (see note).

        Note:
            Make sure all previous commands have passed through the
            compilation chain (call main_engine.flush() to make sure).

        Note:
            If there is a mapper present in the compiler, this function
            automatically converts from logical qubits to mapped qubits for
            the qureg argument.
        """
        qureg = self._convert_logical_to_mapped_qureg(qureg)
        return self._simulator.collapse_wavefunction([qb.id for qb in qureg],
                                                     [bool(int(v)) for v in
                                                      values])

    def cheat(self):
        """
        Access the ordering of the qubits and the state vector directly.

        This is a cheat function which enables, e.g., more efficient
        evaluation of expectation values and debugging.

        Returns:
            A tuple where the first entry is a dictionary mapping qubit
            indices to bit-locations and the second entry is the corresponding
            state vector.

        Note:
            Make sure all previous commands have passed through the
            compilation chain (call main_engine.flush() to make sure).

        Note:
            If there is a mapper present in the compiler, this function
            DOES NOT automatically convert from logical qubits to mapped
            qubits.
        """
        return self._simulator.cheat()

    def _handle(self, cmd):
        """
        Handle all commands, i.e., call the member functions of the C++-
        simulator object corresponding to measurement, allocation/
        deallocation, and (controlled) single-qubit gate.

        Args:
            cmd (Command): Command to handle.

        Raises:
            Exception: If a non-single-qubit gate needs to be processed
                (which should never happen due to is_available).
        """
        if cmd.gate == Measure:
            assert(get_control_count(cmd) == 0)
            ids = [qb.id for qr in cmd.qubits for qb in qr]
            out = self._simulator.measure_qubits(ids)
            i = 0
            for qr in cmd.qubits:
                for qb in qr:
                    # Check if a mapper assigned a different logical id
                    logical_id_tag = None
                    for tag in cmd.tags:
                        if isinstance(tag, LogicalQubitIDTag):
                            logical_id_tag = tag
                    if logical_id_tag is not None:
                        qb = WeakQubitRef(qb.engine,
                                          logical_id_tag.logical_qubit_id)
                    self.main_engine.set_measurement_result(qb, out[i])
                    i += 1
        elif cmd.gate == Allocate:
            ID = cmd.qubits[0][0].id
            self._simulator.allocate_qubit(ID)
        elif cmd.gate == Deallocate:
            ID = cmd.qubits[0][0].id
            self._simulator.deallocate_qubit(ID)
        elif isinstance(cmd.gate, BasicMathGate):
            # improve performance by using C++ code for some commomn gates
            from projectq.libs.math import (AddConstant,
                                            AddConstantModN,
                                            MultiplyByConstantModN)
            qubitids = []
            for qr in cmd.qubits:
                qubitids.append([])
                for qb in qr:
                    qubitids[-1].append(qb.id)
            if FALLBACK_TO_PYSIM:
                math_fun = cmd.gate.get_math_function(cmd.qubits)
                self._simulator.emulate_math(
                    math_fun, qubitids, [qb.id for qb in cmd.control_qubits])
            else:
                # Individual code for different standard gates to make it
                # faster!
                if isinstance(cmd.gate, AddConstant):
                    self._simulator.emulate_math_addConstant(
                        cmd.gate.a, qubitids,
                        [qb.id for qb in cmd.control_qubits])
                elif isinstance(cmd.gate, AddConstantModN):
                    self._simulator.emulate_math_addConstantModN(
                        cmd.gate.a, cmd.gate.N, qubitids,
                        [qb.id for qb in cmd.control_qubits])
                elif isinstance(cmd.gate, MultiplyByConstantModN):
                    self._simulator.emulate_math_multiplyByConstantModN(
                        cmd.gate.a, cmd.gate.N, qubitids,
                        [qb.id for qb in cmd.control_qubits])
                else:
                    math_fun = cmd.gate.get_math_function(cmd.qubits)
                    self._simulator.emulate_math(
                        math_fun, qubitids,
                        [qb.id for qb in cmd.control_qubits])
        elif isinstance(cmd.gate, TimeEvolution):
            op = [(list(term), coeff) for (term, coeff)
                  in cmd.gate.hamiltonian.terms.items()]
            t = cmd.gate.time
            qubitids = [qb.id for qb in cmd.qubits[0]]
            ctrlids = [qb.id for qb in cmd.control_qubits]
            self._simulator.emulate_time_evolution(op, t, qubitids, ctrlids)
        elif len(cmd.gate.matrix) <= 2 ** 5:
            matrix = cmd.gate.matrix
            ids = [qb.id for qr in cmd.qubits for qb in qr]
            if not 2 ** len(ids) == len(cmd.gate.matrix):
                raise Exception("Simulator: Error applying {} gate: "
                                "{}-qubit gate applied to {} qubits.".format(
                                    str(cmd.gate),
                                    int(math.log(len(cmd.gate.matrix), 2)),
                                    len(ids)))
            self._simulator.apply_controlled_gate(matrix.tolist(),
                                                  ids,
                                                  [qb.id for qb in
                                                   cmd.control_qubits])
            if not self._gate_fusion:
                self._simulator.run()
        else:
            raise Exception("This simulator only supports controlled k-qubit"
                            " gates with k < 6!\nPlease add an auto-replacer"
                            " engine to your list of compiler engines.")

    def receive(self, command_list):
        """
        Receive a list of commands from the previous engine and handle them
        (simulate them classically) prior to sending them on to the next
        engine.

        Args:
            command_list (list<Command>): List of commands to execute on the
                simulator.
        """
        for cmd in command_list:
            if not cmd.gate == FlushGate():
                self._handle(cmd)
            else:
                self._simulator.run()  # flush gate --> run all saved gates
            if not self.is_last_engine:
                self.send([cmd])



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
ProjectQ - An open source software framework for quantum computing

Get started:
    Simply import the main compiler engine (from projectq import MainEngine)
    and start coding!

    For examples, see the example folder, which features implementation of
    a quantum random number generator, entanglement demonstration (simulation
    or run on the IBM backend), Teleportation, Grover search, and
    Shor's algorithm for factoring.
"""

# Allow extending this namespace.
__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from ._version import __version__
from projectq.cengines import MainEngine



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Module uses ideas from "Basic circuit compilation techniques
#   for an ion-trap quantum machine" by Dmitri Maslov (2017) at
#   https://iopscience.iop.org/article/10.1088/1367-2630/aa5e47
"""
Apply the restricted gate set setup for trapped ion based quantum computers.

It provides the `engine_list` for the `MainEngine`, restricting the gate set to
Rx and Ry single qubit gates and the Rxx two qubit gates.

A decomposition chooser is implemented following the ideas in QUOTE for
reducing the number of Ry gates in the new circuit.

NOTE:

Because the decomposition chooser is only called when a gate has to be
decomposed, this reduction will work better when the entire circuit has to be
decomposed. Otherwise, If the circuit has both superconding gates and native
ion trapped gates the decomposed circuit will not be optimal.
"""

from projectq.setups import restrictedgateset
from projectq.ops import (Rxx, Rx, Ry)
from projectq.meta import get_control_count

# ------------------chooser_Ry_reducer-------------------#
# If the qubit is not in the prev_Ry_sign dictionary, then no decomposition
# occured
# If the value is:
#  -1 then the last gate applied (during a decomposition!) was Ry(-math.pi/2)
#   1 then the last gate applied (during a decomposition!) was Ry(+math.pi/2)
#   0 then the last gate applied (during a decomposition!) was a Rx

prev_Ry_sign = dict()  # Keeps track of most recent Ry sign, i.e.
#                        whether we had Ry(-pi/2) or Ry(pi/2)
#                        prev_Ry_sign[qubit_index] should hold -1 or
#                        +1


def chooser_Ry_reducer(cmd, decomposition_list):
    """
    Choose the decomposition so as to maximise Ry cancellations, based on the
    previous decomposition used for the given qubit.

    Note:
        Classical instructions gates e.g. Flush and Measure are automatically
        allowed.

    Returns:
        A decomposition object from the decomposition_list.
    """
    decomp_rule = dict()
    name = 'default'

    for decomp in decomposition_list:
        try:
            # NB: need to (possibly) raise an exception before setting the
            # name variable below
            decomposition = decomp.decompose.__name__.split('_')
            decomp_rule[decomposition[3]] = decomp
            name = decomposition[2]
            # 'M' stands for minus, 'P' stands for plus 'N' stands for neutral
            # e.g. decomp_rule['M'] will give you the decomposition_rule that
            # ends with a Ry(-pi/2)
        except IndexError:
            pass

    local_prev_Ry_sign = prev_Ry_sign.setdefault(cmd.engine, dict())

    if name == 'cnot2rxx':
        assert get_control_count(cmd) == 1
        ctrl_id = cmd.control_qubits[0].id

        if local_prev_Ry_sign.get(ctrl_id, -1) <= 0:
            # If the previous qubit had Ry(-pi/2) choose the decomposition
            # that starts with Ry(pi/2)
            local_prev_Ry_sign[ctrl_id] = -1
            # Now the prev_Ry_sign is set to -1 since at the end of the
            # decomposition we will have a Ry(-pi/2)
            return decomp_rule['M']

        # Previous qubit had Ry(pi/2) choose decomposition that starts
        # with Ry(-pi/2) and ends with R(pi/2)
        local_prev_Ry_sign[ctrl_id] = 1
        return decomp_rule['P']

    if name == 'h2rx':
        qubit_id = [qb.id for qureg in cmd.qubits for qb in qureg]
        assert len(qubit_id) == 1  # this should be a single qubit gate
        qubit_id = qubit_id[0]

        if local_prev_Ry_sign.get(qubit_id, 0) == 0:
            local_prev_Ry_sign[qubit_id] = 1
            return decomp_rule['M']

        local_prev_Ry_sign[qubit_id] = 0
        return decomp_rule['N']

    if name == 'rz2rx':
        qubit_id = [qb.id for qureg in cmd.qubits for qb in qureg]
        assert len(qubit_id) == 1  # this should be a single qubit gate
        qubit_id = qubit_id[0]

        if local_prev_Ry_sign.get(qubit_id, -1) <= 0:
            local_prev_Ry_sign[qubit_id] = -1
            return decomp_rule['M']

        local_prev_Ry_sign[qubit_id] = 1
        return decomp_rule['P']

    # No decomposition chosen, so use the first decompostion in the list
    # like the default function
    return decomposition_list[0]


def get_engine_list():
    """
    Returns an engine list compiling code into a trapped ion based compiled
    circuit code.

    Note:

        - Classical instructions gates such as e.g. Flush and Measure are
        automatically allowed.
        - The restricted gate set engine does not work with Rxx gates, as
        ProjectQ will by default bounce back and forth between Cz gates and Cx
        gates. An appropriate decomposition chooser needs to be used!

    Returns:
        A list of suitable compiler engines.
    """
    return restrictedgateset.get_engine_list(
        one_qubit_gates=(Rx, Ry),
        two_qubit_gates=(Rxx, ),
        compiler_chooser=chooser_Ry_reducer)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains a local optimizer engine.
"""

from projectq.cengines import BasicEngine
from projectq.ops import FlushGate, FastForwardingGate, NotMergeable


class LocalOptimizer(BasicEngine):
    """
    LocalOptimizer is a compiler engine which optimizes locally (merging
    rotations, cancelling gates with their inverse) in a local window of user-
    defined size.

    It stores all commands in a dict of lists, where each qubit has its own
    gate pipeline. After adding a gate, it tries to merge / cancel successive
    gates using the get_merged and get_inverse functions of the gate (if
    available). For examples, see BasicRotationGate. Once a list corresponding
    to a qubit contains >=m gates, the pipeline is sent on to the next engine.
    """
    def __init__(self, m=5):
        """
        Initialize a LocalOptimizer object.

        Args:
            m (int): Number of gates to cache per qubit, before sending on the
                first gate.
        """
        BasicEngine.__init__(self)
        self._l = dict()  # dict of lists containing operations for each qubit
        self._m = m  # wait for m gates before sending on

    # sends n gate operations of the qubit with index idx
    def _send_qubit_pipeline(self, idx, n):
        """
        Send n gate operations of the qubit with index idx to the next engine.
        """
        il = self._l[idx]  # temporary label for readability
        for i in range(min(n, len(il))):  # loop over first n operations
            # send all gates before n-qubit gate for other qubits involved
            # --> recursively call send_helper
            other_involved_qubits = [qb
                                     for qreg in il[i].all_qubits
                                     for qb in qreg
                                     if qb.id != idx]
            for qb in other_involved_qubits:
                Id = qb.id
                try:
                    gateloc = 0
                    # find location of this gate within its list
                    while self._l[Id][gateloc] != il[i]:
                        gateloc += 1

                    gateloc = self._optimize(Id, gateloc)

                    # flush the gates before the n-qubit gate
                    self._send_qubit_pipeline(Id, gateloc)
                    # delete the n-qubit gate, we're taking care of it
                    # and don't want the other qubit to do so
                    self._l[Id] = self._l[Id][1:]
                except IndexError:
                    print("Invalid qubit pipeline encountered (in the"
                          " process of shutting down?).")

            # all qubits that need to be flushed have been flushed
            # --> send on the n-qubit gate
            self.send([il[i]])
        # n operations have been sent on --> resize our gate list
        self._l[idx] = self._l[idx][n:]

    def _get_gate_indices(self, idx, i, IDs):
        """
        Return all indices of a command, each index corresponding to the
        command's index in one of the qubits' command lists.

        Args:
            idx (int): qubit index
            i (int): command position in qubit idx's command list
            IDs (list<int>): IDs of all qubits involved in the command
        """
        N = len(IDs)
        # 1-qubit gate: only gate at index i in list #idx is involved
        if N == 1:
            return [i]

        # When the same gate appears multiple time, we need to make sure not to
        # match earlier instances of the gate applied to the same qubits. So we
        # count how many there are, and skip over them when looking in the
        # other lists.
        cmd = self._l[idx][i]
        num_identical_to_skip = sum(1
                                    for prev_cmd in self._l[idx][:i]
                                    if prev_cmd == cmd)
        indices = []
        for Id in IDs:
            identical_indices = [i
                                 for i, c in enumerate(self._l[Id])
                                 if c == cmd]
            indices.append(identical_indices[num_identical_to_skip])
        return indices

    def _optimize(self, idx, lim=None):
        """
        Try to remove identity gates using the is_identity function, then
        merge or even cancel successive gates using the get_merged and
        get_inverse functions of the gate (see, e.g., BasicRotationGate).

        It does so for all qubit command lists.
        """
        # loop over all qubit indices
        i = 0
        limit = len(self._l[idx])
        if lim is not None:
            limit = lim

        while i < limit - 1:
            # can be dropped if the gate is equivalent to an identity gate
            if self._l[idx][i].is_identity():
                # determine index of this gate on all qubits
                qubitids = [qb.id for sublist in self._l[idx][i].all_qubits
                            for qb in sublist]
                gid = self._get_gate_indices(idx, i, qubitids)
                for j in range(len(qubitids)):
                    new_list = (self._l[qubitids[j]][0:gid[j]] +
                                self._l[qubitids[j]][gid[j] + 1:])
                self._l[qubitids[j]] = new_list
                i = 0
                limit -= 1
                continue

            # can be dropped if two in a row are self-inverses
            inv = self._l[idx][i].get_inverse()

            if inv == self._l[idx][i + 1]:
                # determine index of this gate on all qubits
                qubitids = [qb.id for sublist in self._l[idx][i].all_qubits
                            for qb in sublist]
                gid = self._get_gate_indices(idx, i, qubitids)
                # check that there are no other gates between this and its
                # inverse on any of the other qubits involved
                erase = True
                for j in range(len(qubitids)):
                    erase *= (inv == self._l[qubitids[j]][gid[j] + 1])

                # drop these two gates if possible and goto next iteration
                if erase:
                    for j in range(len(qubitids)):
                        new_list = (self._l[qubitids[j]][0:gid[j]] +
                                    self._l[qubitids[j]][gid[j] + 2:])
                        self._l[qubitids[j]] = new_list
                    i = 0
                    limit -= 2
                    continue

            # gates are not each other's inverses --> check if they're
            # mergeable
            try:
                merged_command = self._l[idx][i].get_merged(
                    self._l[idx][i + 1])
                # determine index of this gate on all qubits
                qubitids = [qb.id for sublist in self._l[idx][i].all_qubits
                            for qb in sublist]
                gid = self._get_gate_indices(idx, i, qubitids)

                merge = True
                for j in range(len(qubitids)):
                    m = self._l[qubitids[j]][gid[j]].get_merged(
                        self._l[qubitids[j]][gid[j] + 1])
                    merge *= (m == merged_command)

                if merge:
                    for j in range(len(qubitids)):
                        self._l[qubitids[j]][gid[j]] = merged_command
                        new_list = (self._l[qubitids[j]][0:gid[j] + 1] +
                                    self._l[qubitids[j]][gid[j] + 2:])
                        self._l[qubitids[j]] = new_list
                    i = 0
                    limit -= 1
                    continue
            except NotMergeable:
                pass  # can't merge these two commands.

            i += 1  # next iteration: look at next gate
        return limit

    def _check_and_send(self):
        """
        Check whether a qubit pipeline must be sent on and, if so,
        optimize the pipeline and then send it on.
        """
        for i in self._l:
            if (len(self._l[i]) >= self._m or len(self._l[i]) > 0 and
                    isinstance(self._l[i][-1].gate, FastForwardingGate)):
                self._optimize(i)
                if (len(self._l[i]) >= self._m and not
                        isinstance(self._l[i][-1].gate,
                                   FastForwardingGate)):
                    self._send_qubit_pipeline(i, len(self._l[i]) - self._m + 1)
                elif (len(self._l[i]) > 0 and
                      isinstance(self._l[i][-1].gate, FastForwardingGate)):
                    self._send_qubit_pipeline(i, len(self._l[i]))
        new_dict = dict()
        for idx in self._l:
            if len(self._l[idx]) > 0:
                new_dict[idx] = self._l[idx]
        self._l = new_dict

    def _cache_cmd(self, cmd):
        """
        Cache a command, i.e., inserts it into the command lists of all qubits
        involved.
        """
        # are there qubit ids that haven't been added to the list?
        idlist = [qubit.id for sublist in cmd.all_qubits for qubit in sublist]

        # add gate command to each of the qubits involved
        for ID in idlist:
            if ID not in self._l:
                self._l[ID] = []
            self._l[ID] += [cmd]

        self._check_and_send()

    def receive(self, command_list):
        """
        Receive commands from the previous engine and cache them.
        If a flush gate arrives, the entire buffer is sent on.
        """
        for cmd in command_list:
            if cmd.gate == FlushGate():  # flush gate --> optimize and flush
                for idx in self._l:
                    self._optimize(idx)
                    self._send_qubit_pipeline(idx, len(self._l[idx]))
                new_dict = dict()
                for idx in self._l:
                    if len(self._l[idx]) > 0:
                        new_dict[idx] = self._l[idx]
                self._l = new_dict
                assert self._l == dict()
                self.send([cmd])
            else:
                self._cache_cmd(cmd)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from projectq.ops import BasicGate


class ThisIsNotAGateClassError(TypeError):
    pass


class DecompositionRule:
    """
    A rule for breaking down specific gates into sequences of simpler gates.
    """

    def __init__(self,
                 gate_class,
                 gate_decomposer,
                 gate_recognizer=lambda cmd: True,
                 rule_priority=0):
        """
        Args:
            gate_class (type): The type of gate that this rule decomposes.

                The gate class is redundant information used to make lookups
                faster when iterating over a circuit and deciding "which rules
                apply to this gate?" again and again.

                Note that this parameter is a gate type, not a gate instance.
                You supply gate_class=MyGate or gate_class=MyGate().__class__,
                not gate_class=MyGate().

            gate_decomposer (function[projectq.ops.Command]): Function which,
                given the command to decompose, applies a sequence of gates
                corresponding to the high-level function of a gate of type
                gate_class.

            gate_recognizer (function[projectq.ops.Command] : boolean): A
                predicate that determines if the decomposition applies to the
                given command (on top of the filtering by gate_class).

                For example, a decomposition rule may only to apply rotation
                gates that rotate by a specific angle.

                If no gate_recognizer is given, the decomposition applies to
                all gates matching the gate_class.
        """

        # Check for common gate_class type mistakes.
        if isinstance(gate_class, BasicGate):
            raise ThisIsNotAGateClassError(
                "gate_class is a gate instance instead of a type of BasicGate."
                "\nDid you pass in someGate instead of someGate.__class__?")
        if gate_class == type.__class__:
            raise ThisIsNotAGateClassError(
                "gate_class is type.__class__ instead of a type of BasicGate."
                "\nDid you pass in GateType.__class__ instead of GateType?")

        self.gate_class = gate_class
        self.gate_decomposer = gate_decomposer
        self.gate_recognizer = gate_recognizer
        self.rule_priority = rule_priority



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition to achieve a SqrtSwap gate.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import Compute, Control, Uncompute
from projectq.ops import CNOT, SqrtSwap, SqrtX


def _decompose_sqrtswap(cmd):
    """ Decompose (controlled) swap gates."""
    assert (len(cmd.qubits) == 2 and len(cmd.qubits[0]) == 1 and
            len(cmd.qubits[1]) == 1)
    ctrl = cmd.control_qubits
    qubit0 = cmd.qubits[0][0]
    qubit1 = cmd.qubits[1][0]
    eng = cmd.engine

    with Control(eng, ctrl):
        with Compute(eng):
            CNOT | (qubit0, qubit1)
        with Control(eng, qubit1):
            SqrtX | qubit0
        Uncompute(eng)


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(SqrtSwap.__class__, _decompose_sqrtswap)
]



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import math

from ._basics import ANGLE_PRECISION, ANGLE_TOLERANCE, BasicGate, NotMergeable


class UniformlyControlledRy(BasicGate):
    """
    Uniformly controlled Ry gate as introduced in arXiv:quant-ph/0312218.

    This is an n-qubit gate. There are n-1 control qubits and one target qubit.
    This gate applies Ry(angles(k)) to the target qubit if the n-1 control
    qubits are in the classical state k. As there are 2^(n-1) classical
    states for the control qubits, this gate requires 2^(n-1) (potentially
    different) angle parameters.

    Example:
        .. code-block:: python

        controls = eng.allocate_qureg(2)
        target = eng.allocate_qubit()
        UniformlyControlledRy(angles=[0.1, 0.2, 0.3, 0.4]) | (controls, target)

    Note:
        The first quantum register contains the control qubits. When converting
        the classical state k of the control qubits to an integer, we define
        controls[0] to be the least significant (qu)bit. controls can also
        be an empty list in which case the gate corresponds to an Ry.

    Args:
        angles(list[float]): Rotation angles. Ry(angles[k]) is applied
                             conditioned on the control qubits being in state
                             k.
    """
    def __init__(self, angles):
        BasicGate.__init__(self)
        rounded_angles = []
        for angle in angles:
            new_angle = round(float(angle) % (4. * math.pi), ANGLE_PRECISION)
            if new_angle > 4 * math.pi - ANGLE_TOLERANCE:
                new_angle = 0.
            rounded_angles.append(new_angle)
        self.angles = rounded_angles

    def get_inverse(self):
        return self.__class__([-1 * angle for angle in self.angles])

    def get_merged(self, other):
        if isinstance(other, self.__class__):
            new_angles = [angle1 + angle2 for (angle1, angle2) in
                          zip(self.angles, other.angles)]
            return self.__class__(new_angles)
        raise NotMergeable()

    def __str__(self):
        return "UniformlyControlledRy(" + str(self.angles) + ")"

    def __eq__(self, other):
        """ Return True if same class, same rotation angles."""
        if isinstance(other, self.__class__):
            return self.angles == other.angles
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(str(self))


class UniformlyControlledRz(BasicGate):
    """
    Uniformly controlled Rz gate as introduced in arXiv:quant-ph/0312218.

    This is an n-qubit gate. There are n-1 control qubits and one target qubit.
    This gate applies Rz(angles(k)) to the target qubit if the n-1 control
    qubits are in the classical state k. As there are 2^(n-1) classical
    states for the control qubits, this gate requires 2^(n-1) (potentially
    different) angle parameters.

    Example:
        .. code-block:: python

        controls = eng.allocate_qureg(2)
        target = eng.allocate_qubit()
        UniformlyControlledRz(angles=[0.1, 0.2, 0.3, 0.4]) | (controls, target)

    Note:
        The first quantum register are the contains qubits. When converting
        the classical state k of the control qubits to an integer, we define
        controls[0] to be the least significant (qu)bit. controls can also
        be an empty list in which case the gate corresponds to an Rz.

    Args:
        angles(list[float]): Rotation angles. Rz(angles[k]) is applied
                             conditioned on the control qubits being in state
                             k.
    """
    def __init__(self, angles):
        BasicGate.__init__(self)
        rounded_angles = []
        for angle in angles:
            new_angle = round(float(angle) % (4. * math.pi), ANGLE_PRECISION)
            if new_angle > 4 * math.pi - ANGLE_TOLERANCE:
                new_angle = 0.
            rounded_angles.append(new_angle)
        self.angles = rounded_angles

    def get_inverse(self):
        return self.__class__([-1 * angle for angle in self.angles])

    def get_merged(self, other):
        if isinstance(other, self.__class__):
            new_angles = [angle1 + angle2 for (angle1, angle2) in
                          zip(self.angles, other.angles)]
            return self.__class__(new_angles)
        raise NotMergeable()

    def __str__(self):
        return "UniformlyControlledRz(" + str(self.angles) + ")"

    def __eq__(self, other):
        """ Return True if same class, same rotation angles."""
        if isinstance(other, self.__class__):
            return self.angles == other.angles
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(str(self))



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from projectq.ops import Allocate, Deallocate
from projectq.types import Qubit, Qureg
from projectq.ops import Command


class LastEngineException(Exception):
    """
    Exception thrown when the last engine tries to access the next one.
    (Next engine does not exist)

    The default implementation of isAvailable simply asks the next engine
    whether the command is available. An engine which legally may be the last
    engine, this behavior needs to be adapted (see BasicEngine.isAvailable).
    """
    def __init__(self, engine):
        Exception.__init__(self, ("\nERROR: Sending to next engine failed. "
                                  "{} as last engine?\nIf this is legal, "
                                  "please override 'isAvailable' to adapt its"
                                  " behavior."
                                  ).format(engine.__class__.__name__))


class BasicEngine(object):
    """
    Basic compiler engine: All compiler engines are derived from this class.
    It provides basic functionality such as qubit allocation/deallocation and
    functions that provide information about the engine's position (e.g., next
    engine).

    This information is provided by the MainEngine, which initializes all
    further engines.

    Attributes:
        next_engine (BasicEngine): Next compiler engine (or the back-end).
        main_engine (MainEngine): Reference to the main compiler engine.
        is_last_engine (bool): True for the last engine, which is the back-end.
    """
    def __init__(self):
        """
        Initialize the basic engine.

        Initializes local variables such as _next_engine, _main_engine, etc. to
        None.
        """
        self.main_engine = None
        self.next_engine = None
        self.is_last_engine = False

    def is_available(self, cmd):
        """
        Default implementation of is_available:
        Ask the next engine whether a command is available, i.e.,
        whether it can be executed by the next engine(s).

        Args:
            cmd (Command): Command for which to check availability.

        Returns:
            True if the command can be executed.

        Raises:
            LastEngineException: If is_last_engine is True but is_available
                is not implemented.
        """
        if not self.is_last_engine:
            return self.next_engine.is_available(cmd)
        raise LastEngineException(self)

    def allocate_qubit(self, dirty=False):
        """
        Return a new qubit as a list containing 1 qubit object (quantum
        register of size 1).

        Allocates a new qubit by getting a (new) qubit id from the MainEngine,
        creating the qubit object, and then sending an AllocateQubit command
        down the pipeline. If dirty=True, the fresh qubit can be replaced by
        a pre-allocated one (in an unknown, dirty, initial state). Dirty qubits
        must be returned to their initial states before they are deallocated /
        freed.

        All allocated qubits are added to the MainEngine's set of active
        qubits as weak references. This allows proper clean-up at the end of
        the Python program (using atexit), deallocating all qubits which are
        still alive. Qubit ids of dirty qubits are registered in MainEngine's
        dirty_qubits set.

        Args:
            dirty (bool): If True, indicates that the allocated qubit may be
                dirty (i.e., in an arbitrary initial state).

        Returns:
            Qureg of length 1, where the first entry is the allocated qubit.
        """
        new_id = self.main_engine.get_new_qubit_id()
        qb = Qureg([Qubit(self, new_id)])
        cmd = Command(self, Allocate, (qb,))
        if dirty:
            from projectq.meta import DirtyQubitTag
            if self.is_meta_tag_supported(DirtyQubitTag):
                cmd.tags += [DirtyQubitTag()]
                self.main_engine.dirty_qubits.add(qb[0].id)
        self.main_engine.active_qubits.add(qb[0])
        self.send([cmd])
        return qb

    def allocate_qureg(self, n):
        """
        Allocate n qubits and return them as a quantum register, which is a
        list of qubit objects.

        Args:
            n (int): Number of qubits to allocate
        Returns:
            Qureg of length n, a list of n newly allocated qubits.
        """
        return Qureg([self.allocate_qubit()[0] for _ in range(n)])

    def deallocate_qubit(self, qubit):
        """
        Deallocate a qubit (and sends the deallocation command down the
        pipeline). If the qubit was allocated as a dirty qubit, add
        DirtyQubitTag() to Deallocate command.

        Args:
            qubit (BasicQubit): Qubit to deallocate.
        Raises:
            ValueError: Qubit already deallocated. Caller likely has a bug.
        """
        if qubit.id == -1:
            raise ValueError("Already deallocated.")

        from projectq.meta import DirtyQubitTag
        is_dirty = qubit.id in self.main_engine.dirty_qubits
        self.send([Command(self,
                           Deallocate,
                           (Qureg([qubit]),),
                           tags=[DirtyQubitTag()] if is_dirty else [])])

    def is_meta_tag_supported(self, meta_tag):
        """
        Check if there is a compiler engine handling the meta tag

        Args:
            engine: First engine to check (then iteratively calls
                getNextEngine)
            meta_tag: Meta tag class for which to check support

        Returns:
            supported (bool): True if one of the further compiler engines is a
            meta tag handler, i.e., engine.is_meta_tag_handler(meta_tag)
            returns True.
        """
        engine = self
        try:
            while True:
                try:
                    if engine.is_meta_tag_handler(meta_tag):
                        return True
                except AttributeError:
                    pass
                engine = engine.next_engine
        except AttributeError:
            return False

    def send(self, command_list):
        """
        Forward the list of commands to the next engine in the pipeline.
        """
        self.next_engine.receive(command_list)


class ForwarderEngine(BasicEngine):
    """
    A ForwarderEngine is a trivial engine which forwards all commands to the
    next engine.

    It is mainly used as a substitute for the MainEngine at lower levels such
    that meta operations still work (e.g., with Compute).
    """
    def __init__(self, engine, cmd_mod_fun=None):
        """
        Initialize a ForwarderEngine.

        Args:
            engine (BasicEngine): Engine to forward all commands to.
            cmd_mod_fun (function): Function which is called before sending a
                command. Each command cmd is replaced by the command it
                returns when getting called with cmd.
        """
        BasicEngine.__init__(self)
        self.main_engine = engine.main_engine
        self.next_engine = engine
        if cmd_mod_fun is None:
            def cmd_mod_fun(cmd):
                return cmd

        self._cmd_mod_fun = cmd_mod_fun

    def receive(self, command_list):
        """ Forward all commands to the next engine. """
        new_command_list = [self._cmd_mod_fun(cmd) for cmd in command_list]
        self.send(new_command_list)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition for the Entangle gate.

Applies a Hadamard gate to the first qubit and then, conditioned on this first
qubit, CNOT gates to all others.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import Control
from projectq.ops import X, H, Entangle, All


def _decompose_entangle(cmd):
    """ Decompose the entangle gate. """
    qr = cmd.qubits[0]
    eng = cmd.engine

    with Control(eng, cmd.control_qubits):
        H | qr[0]
        with Control(eng, qr[0]):
            All(X) | qr[1:]


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(Entangle.__class__, _decompose_entangle)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Defines a few shortcuts for certain gates such as
* CNOT = C(NOT)
* CRz = C(Rz)
* Toffoli = C(NOT,2) = C(CNOT)
"""

from ._metagates import C
from ._gates import NOT, Rz, Z


def CRz(angle):
    """
    Shortcut for C(Rz(angle), n=1).
    """
    return C(Rz(angle), n=1)


CNOT = CX = C(NOT)


CZ = C(Z)


Toffoli = C(CNOT)


all_defined_symbols = [CRz]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


"""
Registers decomposition for the TimeEvolution gates.

An exact straight forward decomposition of a TimeEvolution gate is possible
if the hamiltonian has only one term or if all the terms commute with each
other in which case one can implement each term individually.
"""
import math

from projectq.cengines import DecompositionRule
from projectq.meta import Control, Compute, Uncompute
from projectq.ops import TimeEvolution, QubitOperator, H, CNOT, Rz, Rx, Ry


def _recognize_time_evolution_commuting_terms(cmd):
    """
    Recognize all TimeEvolution gates with >1 terms but which all commute.
    """
    hamiltonian = cmd.gate.hamiltonian
    if len(hamiltonian.terms) == 1:
        return False

    id_op = QubitOperator((), 0.0)
    for term in hamiltonian.terms:
        test_op = QubitOperator(term, hamiltonian.terms[term])
        for other in hamiltonian.terms:
            other_op = QubitOperator(other, hamiltonian.terms[other])
            commutator = test_op * other_op - other_op * test_op
            if not commutator.isclose(id_op,
                                      rel_tol=1e-9,
                                      abs_tol=1e-9):
                return False
    return True


def _decompose_time_evolution_commuting_terms(cmd):
    qureg = cmd.qubits
    eng = cmd.engine
    hamiltonian = cmd.gate.hamiltonian
    time = cmd.gate.time
    with Control(eng, cmd.control_qubits):
        for term in hamiltonian.terms:
            ind_operator = QubitOperator(term, hamiltonian.terms[term])
            TimeEvolution(time, ind_operator) | qureg


def _recognize_time_evolution_individual_terms(cmd):
    return len(cmd.gate.hamiltonian.terms) == 1


def _decompose_time_evolution_individual_terms(cmd):
    """
    Implements a TimeEvolution gate with a hamiltonian having only one term.

    To implement exp(-i * t * hamiltonian), where the hamiltonian is only one
    term, e.g., hamiltonian = X0 x Y1 X Z2, we first perform local
    transformations to in order that all Pauli operators in the hamiltonian
    are Z. We then implement  exp(-i * t * (Z1 x Z2 x Z3) and transform the
    basis back to the original. For more details see, e.g.,

    James D. Whitfield, Jacob Biamonte & Aspuru-Guzik
    Simulation of electronic structure Hamiltonians using quantum computers,
    Molecular Physics, 109:5, 735-750 (2011).

    or

    Nielsen and Chuang, Quantum Computation and Information.
    """
    assert len(cmd.qubits) == 1
    qureg = cmd.qubits[0]
    eng = cmd.engine
    time = cmd.gate.time
    hamiltonian = cmd.gate.hamiltonian
    assert len(hamiltonian.terms) == 1
    term = list(hamiltonian.terms)[0]
    coefficient = hamiltonian.terms[term]
    check_indices = set()

    # Check that hamiltonian is not identity term,
    # Previous __or__ operator should have apply a global phase instead:
    assert not term == ()

    # hamiltonian has only a single local operator
    if len(term) == 1:
        with Control(eng, cmd.control_qubits):
            if term[0][1] == 'X':
                Rx(time * coefficient * 2.) | qureg[term[0][0]]
            elif term[0][1] == 'Y':
                Ry(time * coefficient * 2.) | qureg[term[0][0]]
            else:
                Rz(time * coefficient * 2.) | qureg[term[0][0]]
    # hamiltonian has more than one local operator
    else:
        with Control(eng, cmd.control_qubits):
            with Compute(eng):
                # Apply local basis rotations
                for index, action in term:
                    check_indices.add(index)
                    if action == 'X':
                        H | qureg[index]
                    elif action == 'Y':
                        Rx(math.pi / 2.) | qureg[index]
                # Check that qureg had exactly as many qubits as indices:
                assert check_indices == set((range(len(qureg))))
                # Compute parity
                for i in range(len(qureg)-1):
                    CNOT | (qureg[i], qureg[i+1])
            Rz(time * coefficient * 2.) | qureg[-1]
            # Uncompute parity and basis change
            Uncompute(eng)


rule_commuting_terms = DecompositionRule(
    gate_class=TimeEvolution,
    gate_decomposer=_decompose_time_evolution_commuting_terms,
    gate_recognizer=_recognize_time_evolution_commuting_terms)


rule_individual_terms = DecompositionRule(
    gate_class=TimeEvolution,
    gate_decomposer=_decompose_time_evolution_individual_terms,
    gate_recognizer=_recognize_time_evolution_individual_terms)


#: Decomposition rules
all_defined_decomposition_rules = [rule_commuting_terms,
                                   rule_individual_terms]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""QubitOperator stores a sum of Pauli operators acting on qubits."""
import cmath
import copy

from ._basics import BasicGate, NotInvertible, NotMergeable
from ._command import apply_command
from ._gates import Ph, X, Y, Z

EQ_TOLERANCE = 1e-12

# Define products of all Pauli operators for symbolic multiplication.
_PAULI_OPERATOR_PRODUCTS = {
    ('I', 'I'): (1., 'I'),
    ('I', 'X'): (1., 'X'),
    ('X', 'I'): (1., 'X'),
    ('I', 'Y'): (1., 'Y'),
    ('Y', 'I'): (1., 'Y'),
    ('I', 'Z'): (1., 'Z'),
    ('Z', 'I'): (1., 'Z'),
    ('X', 'X'): (1., 'I'),
    ('Y', 'Y'): (1., 'I'),
    ('Z', 'Z'): (1., 'I'),
    ('X', 'Y'): (1.j, 'Z'),
    ('X', 'Z'): (-1.j, 'Y'),
    ('Y', 'X'): (-1.j, 'Z'),
    ('Y', 'Z'): (1.j, 'X'),
    ('Z', 'X'): (1.j, 'Y'),
    ('Z', 'Y'): (-1.j, 'X')
}


class QubitOperatorError(Exception):
    pass


class QubitOperator(BasicGate):
    """
    A sum of terms acting on qubits, e.g., 0.5 * 'X0 X5' + 0.3 * 'Z1 Z2'.

    A term is an operator acting on n qubits and can be represented as:

    coefficent * local_operator[0] x ... x local_operator[n-1]

    where x is the tensor product. A local operator is a Pauli operator
    ('I', 'X', 'Y', or 'Z') which acts on one qubit. In math notation a term
    is, for example, 0.5 * 'X0 X5', which means that a Pauli X operator acts
    on qubit 0 and 5, while the identity operator acts on all other qubits.

    A QubitOperator represents a sum of terms acting on qubits and overloads
    operations for easy manipulation of these objects by the user.

    Note for a QubitOperator to be a Hamiltonian which is a hermitian
    operator, the coefficients of all terms must be real.

    .. code-block:: python

        hamiltonian = 0.5 * QubitOperator('X0 X5') + 0.3 * QubitOperator('Z0')

    Our Simulator takes a hermitian QubitOperator to directly calculate the
    expectation value (see Simulator.get_expectation_value) of this observable.

    A hermitian QubitOperator can also be used as input for the
    TimeEvolution gate.

    If the QubitOperator is unitary, i.e., it contains only one term with a
    coefficient, whose absolute value is 1, then one can apply it directly to
    qubits:

    .. code-block:: python

        eng = projectq.MainEngine()
        qureg = eng.allocate_qureg(6)
        QubitOperator('X0 X5', 1.j) | qureg  # Applies X to qubit 0 and 5
                                             # with an additional global phase
                                             # of 1.j


    Attributes:
        terms (dict): **key**: A term represented by a tuple containing all
                      non-trivial local Pauli operators ('X', 'Y', or 'Z').
                      A non-trivial local Pauli operator is specified by a
                      tuple with the first element being an integer
                      indicating the qubit on which a non-trivial local
                      operator acts and the second element being a string,
                      either 'X', 'Y', or 'Z', indicating which non-trivial
                      Pauli operator acts on that qubit. Examples:
                      ((1, 'X'),) or ((1, 'X'), (4,'Z')) or the identity ().
                      The tuples representing the non-trivial local terms
                      are sorted according to the qubit number they act on,
                      starting from 0.
                      **value**: Coefficient of this term as a (complex) float
    """

    def __init__(self, term=None, coefficient=1.):
        """
        Inits a QubitOperator.

        The init function only allows to initialize one term. Additional terms
        have to be added using += (which is fast) or using + of two
        QubitOperator objects:

        Example:
            .. code-block:: python

                ham = ((QubitOperator('X0 Y3', 0.5)
                        + 0.6 * QubitOperator('X0 Y3')))
                # Equivalently
                ham2 = QubitOperator('X0 Y3', 0.5)
                ham2 += 0.6 * QubitOperator('X0 Y3')

        Note:
            Adding terms to QubitOperator is faster using += (as this is done
            by in-place addition). Specifying the coefficient in the __init__
            is faster than by multiplying a QubitOperator with a scalar as
            calls an out-of-place multiplication.

        Args:
            coefficient (complex float, optional): The coefficient of the
                first term of this QubitOperator. Default is 1.0.
            term (optional, empy tuple, a tuple of tuples, or a string):
                1) Default is None which means there are no terms in the
                   QubitOperator hence it is the "zero" Operator
                2) An empty tuple means there are no non-trivial Pauli
                   operators acting on the qubits hence only identities
                   with a coefficient (which by default is 1.0).
                3) A sorted tuple of tuples. The first element of each tuple
                   is an integer indicating the qubit on which a non-trivial
                   local operator acts, starting from zero. The second element
                   of each tuple is a string, either 'X', 'Y' or 'Z',
                   indicating which local operator acts on that qubit.
                4) A string of the form 'X0 Z2 Y5', indicating an X on
                   qubit 0, Z on qubit 2, and Y on qubit 5. The string should
                   be sorted by the qubit number. '' is the identity.

        Raises:
            QubitOperatorError: Invalid operators provided to QubitOperator.
        """
        BasicGate.__init__(self)
        if not isinstance(coefficient, (int, float, complex)):
            raise ValueError('Coefficient must be a numeric type.')
        self.terms = {}
        if term is None:
            return
        if isinstance(term, tuple):
            if term is ():
                self.terms[()] = coefficient
            else:
                # Test that input is a tuple of tuples and correct action
                for local_operator in term:
                    if (not isinstance(local_operator, tuple)
                            or len(local_operator) != 2):
                        raise ValueError("term specified incorrectly.")
                    qubit_num, action = local_operator
                    if not isinstance(action, str) or action not in 'XYZ':
                        raise ValueError("Invalid action provided: must be "
                                         "string 'X', 'Y', or 'Z'.")
                    if not (isinstance(qubit_num, int) and qubit_num >= 0):
                        raise QubitOperatorError("Invalid qubit number "
                                                 "provided to QubitTerm: "
                                                 "must be a non-negative "
                                                 "int.")
                # Sort and add to self.terms:
                term = list(term)
                term.sort(key=lambda loc_operator: loc_operator[0])
                self.terms[tuple(term)] = coefficient
        elif isinstance(term, str):
            list_ops = []
            for el in term.split():
                if len(el) < 2:
                    raise ValueError('term specified incorrectly.')
                list_ops.append((int(el[1:]), el[0]))
            # Test that list_ops has correct format of tuples
            for local_operator in list_ops:
                qubit_num, action = local_operator
                if not isinstance(action, str) or action not in 'XYZ':
                    raise ValueError("Invalid action provided: must be "
                                     "string 'X', 'Y', or 'Z'.")
                if not (isinstance(qubit_num, int) and qubit_num >= 0):
                    raise QubitOperatorError("Invalid qubit number "
                                             "provided to QubitTerm: "
                                             "must be a non-negative "
                                             "int.")
            # Sort and add to self.terms:
            list_ops.sort(key=lambda loc_operator: loc_operator[0])
            self.terms[tuple(list_ops)] = coefficient
        else:
            raise ValueError('term specified incorrectly.')

    def compress(self, abs_tol=1e-12):
        """
        Eliminates all terms with coefficients close to zero and removes
        imaginary parts of coefficients that are close to zero.

        Args:
            abs_tol(float): Absolute tolerance, must be at least 0.0
        """
        new_terms = {}
        for term in self.terms:
            coeff = self.terms[term]
            if abs(coeff.imag) <= abs_tol:
                coeff = coeff.real
            if abs(coeff) > abs_tol:
                new_terms[term] = coeff
        self.terms = new_terms

    def isclose(self, other, rel_tol=1e-12, abs_tol=1e-12):
        """
        Returns True if other (QubitOperator) is close to self.

        Comparison is done for each term individually. Return True
        if the difference between each term in self and other is
        less than the relative tolerance w.r.t. either other or self
        (symmetric test) or if the difference is less than the absolute
        tolerance.

        Args:
            other(QubitOperator): QubitOperator to compare against.
            rel_tol(float): Relative tolerance, must be greater than 0.0
            abs_tol(float): Absolute tolerance, must be at least 0.0
        """
        # terms which are in both:
        for term in set(self.terms).intersection(set(other.terms)):
            a = self.terms[term]
            b = other.terms[term]
            # math.isclose does this in Python >=3.5
            if not abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol):
                return False
        # terms only in one (compare to 0.0 so only abs_tol)
        for term in set(self.terms).symmetric_difference(set(other.terms)):
            if term in self.terms:
                if not abs(self.terms[term]) <= abs_tol:
                    return False
            elif not abs(other.terms[term]) <= abs_tol:
                return False
        return True

    def __or__(self, qubits):
        """
        Operator| overload which enables the following syntax:

        .. code-block:: python

            QubitOperator(...) | qureg
            QubitOperator(...) | (qureg,)
            QubitOperator(...) | qubit
            QubitOperator(...) | (qubit,)

        Unlike other gates, this gate is only allowed to be applied to one
        quantum register or one qubit and only if the QubitOperator is
        unitary, i.e., consists of one term with a coefficient whose absolute
        values is 1.

        Example:

        .. code-block:: python

            eng = projectq.MainEngine()
            qureg = eng.allocate_qureg(6)
            QubitOperator('X0 X5', 1.j) | qureg  # Applies X to qubit 0 and 5
                                                 # with an additional global
                                                 # phase of 1.j

        While in the above example the QubitOperator gate is applied to 6
        qubits, it only acts non-trivially on the two qubits qureg[0] and
        qureg[5]. Therefore, the operator| will create a new rescaled
        QubitOperator, i.e, it sends the equivalent of the following new gate
        to the MainEngine:

        .. code-block:: python

            QubitOperator('X0 X1', 1.j) | [qureg[0], qureg[5]]

        which is only a two qubit gate.

        Args:
            qubits: one Qubit object, one list of Qubit objects, one Qureg
                    object, or a tuple of the former three cases.

        Raises:
            TypeError: If QubitOperator is not unitary or applied to more than
                       one quantum register.
            ValueError: If quantum register does not have enough qubits
        """
        # Check that input is only one qureg or one qubit
        qubits = self.make_tuple_of_qureg(qubits)
        if len(qubits) != 1:
            raise TypeError("Only one qubit or qureg allowed.")
        # Check that operator is unitary
        if not len(self.terms) == 1:
            raise TypeError("Too many terms. Only QubitOperators consisting "
                            "of a single term (single n-qubit Pauli operator) "
                            "with a coefficient of unit length can be applied "
                            "to qubits with this function.")
        (term, coefficient), = self.terms.items()
        phase = cmath.phase(coefficient)
        if (abs(coefficient) < 1 - EQ_TOLERANCE
                or abs(coefficient) > 1 + EQ_TOLERANCE):
            raise TypeError("abs(coefficient) != 1. Only QubitOperators "
                            "consisting of a single term (single n-qubit "
                            "Pauli operator) with a coefficient of unit "
                            "length can be applied to qubits with this "
                            "function.")
        # Test if we need to apply only Ph
        if term == ():
            Ph(phase) | qubits[0][0]
            return
        # Check that Qureg has enough qubits:
        num_qubits = len(qubits[0])
        non_trivial_qubits = set()
        for index, action in term:
            non_trivial_qubits.add(index)
        if max(non_trivial_qubits) >= num_qubits:
            raise ValueError("QubitOperator acts on more qubits than the gate "
                             "is applied to.")
        # Apply X, Y, Z, if QubitOperator acts only on one qubit
        if len(term) == 1:
            if term[0][1] == "X":
                X | qubits[0][term[0][0]]
            elif term[0][1] == "Y":
                Y | qubits[0][term[0][0]]
            elif term[0][1] == "Z":
                Z | qubits[0][term[0][0]]
            Ph(phase) | qubits[0][term[0][0]]
            return
        # Create new QubitOperator gate with rescaled qubit indices in
        # 0,..., len(non_trivial_qubits) - 1
        new_index = dict()
        non_trivial_qubits = sorted(list(non_trivial_qubits))
        for i in range(len(non_trivial_qubits)):
            new_index[non_trivial_qubits[i]] = i
        new_qubitoperator = QubitOperator()
        assert len(new_qubitoperator.terms) == 0
        new_term = tuple(
            [(new_index[index], action) for index, action in term])
        new_qubitoperator.terms[new_term] = coefficient
        new_qubits = [qubits[0][i] for i in non_trivial_qubits]
        # Apply new gate
        cmd = new_qubitoperator.generate_command(new_qubits)
        apply_command(cmd)

    def get_inverse(self):
        """
        Return the inverse gate of a QubitOperator if applied as a gate.

        Raises:
            NotInvertible: Not implemented for QubitOperators which have
                           multiple terms or a coefficient with absolute value
                           not equal to 1.
        """

        if len(self.terms) == 1:
            (term, coefficient), = self.terms.items()
            if (not abs(coefficient) < 1 - EQ_TOLERANCE
                    and not abs(coefficient) > 1 + EQ_TOLERANCE):
                return QubitOperator(term, coefficient**(-1))
        raise NotInvertible("BasicGate: No get_inverse() implemented.")

    def get_merged(self, other):
        """
        Return this gate merged with another gate.

        Standard implementation of get_merged:

        Raises:
            NotMergeable: merging is not possible
        """
        if (isinstance(other, self.__class__) and len(other.terms) == 1
                and len(self.terms) == 1):
            return self * other
        raise NotMergeable()

    def __imul__(self, multiplier):
        """
        In-place multiply (*=) terms with scalar or QubitOperator.

        Args:
            multiplier(complex float, or QubitOperator): multiplier
        """
        # Handle scalars.
        if isinstance(multiplier, (int, float, complex)):
            for term in self.terms:
                self.terms[term] *= multiplier
            return self

        # Handle QubitOperator.
        if isinstance(multiplier, QubitOperator):
            result_terms = dict()
            for left_term in self.terms:
                for right_term in multiplier.terms:
                    new_coefficient = (
                        self.terms[left_term] * multiplier.terms[right_term])

                    # Loop through local operators and create new sorted list
                    # of representing the product local operator:
                    product_operators = []
                    left_operator_index = 0
                    right_operator_index = 0
                    n_operators_left = len(left_term)
                    n_operators_right = len(right_term)
                    while (left_operator_index < n_operators_left
                           and right_operator_index < n_operators_right):
                        (left_qubit,
                         left_loc_op) = (left_term[left_operator_index])
                        (right_qubit,
                         right_loc_op) = (right_term[right_operator_index])

                        # Multiply local operators acting on the same qubit
                        if left_qubit == right_qubit:
                            left_operator_index += 1
                            right_operator_index += 1
                            (scalar,
                             loc_op) = _PAULI_OPERATOR_PRODUCTS[(left_loc_op,
                                                                 right_loc_op)]

                            # Add new term.
                            if loc_op != 'I':
                                product_operators += [(left_qubit, loc_op)]
                                new_coefficient *= scalar
                            # Note if loc_op == 'I', then scalar == 1.0

                        # If left_qubit > right_qubit, add right_loc_op; else,
                        # add left_loc_op.
                        elif left_qubit > right_qubit:
                            product_operators += [(right_qubit, right_loc_op)]
                            right_operator_index += 1
                        else:
                            product_operators += [(left_qubit, left_loc_op)]
                            left_operator_index += 1

                    # Finish the remainding operators:
                    if left_operator_index == n_operators_left:
                        product_operators += right_term[right_operator_index::]
                    elif right_operator_index == n_operators_right:
                        product_operators += left_term[left_operator_index::]

                    # Add to result dict
                    tmp_key = tuple(product_operators)
                    if tmp_key in result_terms:
                        result_terms[tmp_key] += new_coefficient
                    else:
                        result_terms[tmp_key] = new_coefficient
            self.terms = result_terms
            return self

        raise TypeError('Cannot in-place multiply term of invalid type ' +
                        'to QubitTerm.')

    def __mul__(self, multiplier):
        """
        Return self * multiplier for a scalar, or a QubitOperator.

        Args:
            multiplier: A scalar, or a QubitOperator.

        Returns:
            product: A QubitOperator.

        Raises:
            TypeError: Invalid type cannot be multiply with QubitOperator.
        """
        if isinstance(multiplier, (int, float, complex, QubitOperator)):
            product = copy.deepcopy(self)
            product *= multiplier
            return product

        raise TypeError(
            'Object of invalid type cannot multiply with QubitOperator.')

    def __rmul__(self, multiplier):
        """
        Return multiplier * self for a scalar.

        We only define __rmul__ for scalars because the left multiply
        exist for  QubitOperator and left multiply
        is also queried as the default behavior.

        Args:
            multiplier: A scalar to multiply by.

        Returns:
            product: A new instance of QubitOperator.

        Raises:
            TypeError: Object of invalid type cannot multiply QubitOperator.
        """
        if not isinstance(multiplier, (int, float, complex)):
            raise TypeError(
                'Object of invalid type cannot multiply with QubitOperator.')
        return self * multiplier

    def __truediv__(self, divisor):
        """
        Return self / divisor for a scalar.

        Note:
            This is always floating point division.

        Args:
            divisor: A scalar to divide by.

        Returns:
            A new instance of QubitOperator.

        Raises:
            TypeError: Cannot divide local operator by non-scalar type.
        """
        if not isinstance(divisor, (int, float, complex)):
            raise TypeError('Cannot divide QubitOperator by non-scalar type.')
        return self * (1.0 / divisor)

    def __div__(self, divisor):
        """ For compatibility with Python 2. """
        return self.__truediv__(divisor)

    def __itruediv__(self, divisor):
        if not isinstance(divisor, (int, float, complex)):
            raise TypeError('Cannot divide QubitOperator by non-scalar type.')
        self *= (1.0 / divisor)
        return self

    def __idiv__(self, divisor):
        """ For compatibility with Python 2. """
        return self.__itruediv__(divisor)

    def __iadd__(self, addend):
        """
        In-place method for += addition of QubitOperator.

        Args:
            addend: A QubitOperator.

        Raises:
            TypeError: Cannot add invalid type.
        """
        if isinstance(addend, QubitOperator):
            for term in addend.terms:
                if term in self.terms:
                    if abs(addend.terms[term] + self.terms[term]) > 0.:
                        self.terms[term] += addend.terms[term]
                    else:
                        self.terms.pop(term)
                else:
                    self.terms[term] = addend.terms[term]
        else:
            raise TypeError('Cannot add invalid type to QubitOperator.')
        return self

    def __add__(self, addend):
        """ Return self + addend for a QubitOperator. """
        summand = copy.deepcopy(self)
        summand += addend
        return summand

    def __isub__(self, subtrahend):
        """
        In-place method for -= subtraction of QubitOperator.

        Args:
            subtrahend: A QubitOperator.

        Raises:
            TypeError: Cannot subtract invalid type from QubitOperator.
        """
        if isinstance(subtrahend, QubitOperator):
            for term in subtrahend.terms:
                if term in self.terms:
                    if abs(self.terms[term] - subtrahend.terms[term]) > 0.:
                        self.terms[term] -= subtrahend.terms[term]
                    else:
                        self.terms.pop(term)
                else:
                    self.terms[term] = -subtrahend.terms[term]
        else:
            raise TypeError('Cannot subtract invalid type from QubitOperator.')
        return self

    def __sub__(self, subtrahend):
        """ Return self - subtrahend for a QubitOperator. """
        minuend = copy.deepcopy(self)
        minuend -= subtrahend
        return minuend

    def __neg__(self):
        return -1. * self

    def __str__(self):
        """Return an easy-to-read string representation."""
        if not self.terms:
            return '0'
        string_rep = ''
        for term in self.terms:
            tmp_string = '{}'.format(self.terms[term])
            if term == ():
                tmp_string += ' I'
            for operator in term:
                if operator[1] == 'X':
                    tmp_string += ' X{}'.format(operator[0])
                elif operator[1] == 'Y':
                    tmp_string += ' Y{}'.format(operator[0])
                else:
                    assert operator[1] == 'Z'
                    tmp_string += ' Z{}'.format(operator[0])
            string_rep += '{} +\n'.format(tmp_string)
        return string_rep[:-3]

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self))



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
This module provides the basic functionality required to plot a quantum
circuit in a matplotlib figure.
It is mainly used by the CircuitDrawerMatplotlib compiler engine.

Currently, it supports all single-qubit gates, including their controlled
versions to an arbitrary number of control qubits. It also supports
multi-target qubit gates under some restrictions. Namely that the target
qubits must be neighbours in the output figure (which cannot be determined
durinng compilation at this time).
"""

from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection, LineCollection
from matplotlib.lines import Line2D
from matplotlib.patches import Circle, Arc, Rectangle

# Important note on units for the plot parameters.
# The following entries are in inches:
#   - column_spacing
#   - labels_margin
#   - wire_height
#
# The following entries are in data units (matplotlib)
#   - control_radius
#   - gate_offset
#   - mgate_width
#   - not_radius
#   - swap_delta
#   - x_offset
#
# The rest have misc. units (as defined by matplotlib)
_DEFAULT_PLOT_PARAMS = dict(fontsize=14.0,
                            column_spacing=.5,
                            control_radius=0.015,
                            labels_margin=1,
                            linewidth=1.0,
                            not_radius=0.03,
                            gate_offset=.05,
                            mgate_width=0.1,
                            swap_delta=0.02,
                            x_offset=.05,
                            wire_height=1)

# ==============================================================================


def to_draw(qubit_lines, qubit_labels=None, drawing_order=None, **kwargs):
    """
    Translates a given circuit to a matplotlib figure.

    Args:
        qubit_lines (dict): list of gates for each qubit axis
        qubit_labels (dict): label to print in front of the qubit wire for
            each qubit ID
        drawing_order (dict): index of the wire for each qubit ID to be drawn.
        **kwargs (dict): additional parameters are used to update the default
            plot parameters

    Returns:
        A tuple with (figure, axes)

    Note:
        Numbering of qubit wires starts at 0 at the bottom and increases
        vertically.

    Note:
        Additional keyword arguments can be passed to this
        function in order to further customize the figure output
        by matplotlib (default value in parentheses):

          - fontsize (14): Font size in pt
          - column_spacing (.5): Vertical spacing between two
            neighbouring gates (roughly in inches)
          - control_radius (.015): Radius of the circle for controls
          - labels_margin (1): Margin between labels and begin of
            wire (roughly in inches)
          - linewidth (1): Width of line
          - not_radius (.03): Radius of the circle for X/NOT gates
          - gate_offset (.05): Inner margins for gates with a text
            representation
          - mgate_width (.1): Width of the measurement gate
          - swap_delta (.02): Half-size of the SWAP gate
          - x_offset (.05): Absolute X-offset for drawing within the axes
          - wire_height (1): Vertical spacing between two qubit
            wires (roughly in inches)
    """
    if qubit_labels is None:
        qubit_labels = {qubit_id: r'$|0\rangle$' for qubit_id in qubit_lines}
    else:
        if list(qubit_labels) != list(qubit_lines):
            raise RuntimeError('Qubit IDs in qubit_labels do not match '
                               + 'qubit IDs in qubit_lines!')

    if drawing_order is None:
        n_qubits = len(qubit_lines)
        drawing_order = {
            qubit_id: n_qubits - qubit_id - 1
            for qubit_id in list(qubit_lines)
        }
    else:
        if list(drawing_order) != list(qubit_lines):
            raise RuntimeError('Qubit IDs in drawing_order do not match '
                               + 'qubit IDs in qubit_lines!')
        if (list(sorted(drawing_order.values())) != list(
                range(len(drawing_order)))):
            raise RuntimeError(
                'Indices of qubit wires in drawing_order '
                + 'must be between 0 and {}!'.format(len(drawing_order)))

    plot_params = deepcopy(_DEFAULT_PLOT_PARAMS)
    plot_params.update(kwargs)

    n_labels = len(list(qubit_lines))

    wire_height = plot_params['wire_height']
    # Grid in inches
    wire_grid = np.arange(wire_height, (n_labels + 1) * wire_height,
                          wire_height,
                          dtype=float)

    fig, axes = create_figure(plot_params)

    # Grid in inches
    gate_grid = calculate_gate_grid(axes, qubit_lines, plot_params)

    width = gate_grid[-1] + plot_params['column_spacing']
    height = wire_grid[-1] + wire_height

    resize_figure(fig, axes, width, height, plot_params)

    # Convert grids into data coordinates
    units_per_inch = plot_params['units_per_inch']

    gate_grid *= units_per_inch
    gate_grid = gate_grid + plot_params['x_offset']
    wire_grid *= units_per_inch
    plot_params['column_spacing'] *= units_per_inch

    draw_wires(axes, n_labels, gate_grid, wire_grid, plot_params)

    draw_labels(axes, qubit_labels, drawing_order, wire_grid, plot_params)

    draw_gates(axes, qubit_lines, drawing_order, gate_grid, wire_grid,
               plot_params)
    return fig, axes


# ==============================================================================
# Functions used to calculate the layout


def gate_width(axes, gate_str, plot_params):
    """
    Calculate the width of a gate based on its string representation.

    Args:
        axes (matplotlib.axes.Axes): axes object
        gate_str (str): string representation of a gate
        plot_params (dict): plot parameters

    Returns:
        The width of a gate on the figure (in inches)
    """
    if gate_str == 'X':
        return 2 * plot_params['not_radius'] / plot_params['units_per_inch']
    if gate_str == 'Swap':
        return 2 * plot_params['swap_delta'] / plot_params['units_per_inch']

    if gate_str == 'Measure':
        return plot_params['mgate_width']

    obj = axes.text(0,
                    0,
                    gate_str,
                    visible=True,
                    bbox=dict(edgecolor='k', facecolor='w', fill=True, lw=1.0),
                    fontsize=14)
    obj.figure.canvas.draw()
    width = (obj.get_window_extent(obj.figure.canvas.get_renderer()).width
             / axes.figure.dpi)
    obj.remove()
    return width + 2 * plot_params['gate_offset']


def calculate_gate_grid(axes, qubit_lines, plot_params):
    """
    Calculate an optimal grid spacing for a list of quantum gates.

    Args:
        axes (matplotlib.axes.Axes): axes object
        qubit_lines (dict): list of gates for each qubit axis
        plot_params (dict): plot parameters

    Returns:
        An array (np.ndarray) with the gate x positions.
    """
    # NB: column_spacing is still in inch when this function is called
    column_spacing = plot_params['column_spacing']
    data = list(qubit_lines.values())
    depth = len(data[0])

    width_list = [
        max(
            gate_width(axes, line[idx][0], plot_params) if line[idx] else 0
            for line in data) for idx in range(depth)
    ]

    gate_grid = np.array([0] * (depth + 1), dtype=float)
    
    gate_grid[0] = plot_params['labels_margin']
    if depth > 0:
        gate_grid[0] += width_list[0] * 0.5
        for idx in range(1, depth):
            gate_grid[idx] = gate_grid[idx - 1] + column_spacing + (
                width_list[idx] + width_list[idx - 1]) * 0.5
        gate_grid[-1] = gate_grid[-2] + column_spacing + width_list[-1] * 0.5
    return gate_grid


# ==============================================================================
# Basic helper functions


def text(axes, gate_pos, wire_pos, textstr, plot_params):
    """
    Draws a text box on the figure.

    Args:
        axes (matplotlib.axes.Axes): axes object
        gate_pos (float): x coordinate of the gate [data units]
        wire_pos (float): y coordinate of the qubit wire
        textstr (str): text of the gate and box
        plot_params (dict): plot parameters
        box (bool): draw the rectangle box if box is True
    """
    return axes.text(gate_pos,
                     wire_pos,
                     textstr,
                     color='k',
                     ha='center',
                     va='center',
                     clip_on=True,
                     size=plot_params['fontsize'])


# ==============================================================================


def create_figure(plot_params):
    """
    Create a new figure as well as a new axes instance

    Args:
        plot_params (dict): plot parameters

    Returns:
        A tuple with (figure, axes)
    """
    fig = plt.figure(facecolor='w', edgecolor='w')
    axes = plt.axes()
    axes.set_axis_off()
    axes.set_aspect('equal')
    plot_params['units_per_inch'] = fig.dpi / axes.get_window_extent().width
    return fig, axes


def resize_figure(fig, axes, width, height, plot_params):
    """
    Resizes a figure and adjust the limits of the axes instance to make sure
    that the distances in data coordinates on the screen stay constant.

    Args:
        fig (matplotlib.figure.Figure): figure object
        axes (matplotlib.axes.Axes): axes object
        width (float): new figure width
        height (float): new figure height
        plot_params (dict): plot parameters

    Returns:
        A tuple with (figure, axes)
    """
    fig.set_size_inches(width, height)

    new_limits = plot_params['units_per_inch'] * np.array([width, height])
    axes.set_xlim(0, new_limits[0])
    axes.set_ylim(0, new_limits[1])


def draw_gates(axes, qubit_lines, drawing_order, gate_grid, wire_grid,
               plot_params):
    """
    Draws the gates.

    Args:
        qubit_lines (dict): list of gates for each qubit axis
        drawing_order (dict): index of the wire for each qubit ID to be drawn
        gate_grid (np.ndarray): x positions of the gates
        wire_grid (np.ndarray): y positions of the qubit wires
        plot_params (dict): plot parameters

    Returns:
        A tuple with (figure, axes)
    """
    for qubit_line in qubit_lines.values():
        for idx, data in enumerate(qubit_line):
            if data is not None:
                (gate_str, targets, controls) = data
                targets_order = [drawing_order[tgt] for tgt in targets]
                draw_gate(
                    axes, gate_str, gate_grid[idx],
                    [wire_grid[tgt] for tgt in targets_order], targets_order,
                    [wire_grid[drawing_order[ctrl]]
                     for ctrl in controls], plot_params)


def draw_gate(axes, gate_str, gate_pos, target_wires, targets_order,
              control_wires, plot_params):
    """
    Draws a single gate at a given location.

    Args:
        axes (AxesSubplot): axes object
        gate_str (str): string representation of a gate
        gate_pos (float): x coordinate of the gate [data units]
        target_wires (list): y coordinates of the target qubits
        targets_order (list): index of the wires corresponding to the target
                              qubit IDs
        control_wires (list): y coordinates of the control qubits
        plot_params (dict): plot parameters

    Returns:
        A tuple with (figure, axes)
    """
    # Special cases
    if gate_str == 'Z' and len(control_wires) == 1:
        draw_control_z_gate(axes, gate_pos, target_wires[0], control_wires[0],
                            plot_params)
    elif gate_str == 'X':
        draw_x_gate(axes, gate_pos, target_wires[0], plot_params)
    elif gate_str == 'Swap':
        draw_swap_gate(axes, gate_pos, target_wires[0], target_wires[1],
                       plot_params)
    elif gate_str == 'Measure':
        draw_measure_gate(axes, gate_pos, target_wires[0], plot_params)
    else:
        if len(target_wires) == 1:
            draw_generic_gate(axes, gate_pos, target_wires[0], gate_str,
                              plot_params)
        else:
            if sorted(targets_order) != list(
                    range(min(targets_order),
                          max(targets_order) + 1)):
                raise RuntimeError(
                    'Multi-qubit gate with non-neighbouring qubits!\n'
                    + 'Gate: {} on wires {}'.format(gate_str, targets_order))

            multi_qubit_gate(axes, gate_str, gate_pos, min(target_wires),
                             max(target_wires), plot_params)

    if not control_wires:
        return

    for control_wire in control_wires:
        axes.add_patch(
            Circle((gate_pos, control_wire),
                   plot_params['control_radius'],
                   ec='k',
                   fc='k',
                   fill=True,
                   lw=plot_params['linewidth']))

    all_wires = target_wires + control_wires
    axes.add_line(
        Line2D((gate_pos, gate_pos), (min(all_wires), max(all_wires)),
               color='k',
               lw=plot_params['linewidth']))


def draw_generic_gate(axes, gate_pos, wire_pos, gate_str, plot_params):
    """
    Draws a measurement gate.

    Args:
        axes (AxesSubplot): axes object
        gate_pos (float): x coordinate of the gate [data units]
        wire_pos (float): y coordinate of the qubit wire
        gate_str (str) : string representation of a gate
        plot_params (dict): plot parameters
    """
    obj = text(axes, gate_pos, wire_pos, gate_str, plot_params)
    obj.set_zorder(7)

    factor = plot_params['units_per_inch'] / obj.figure.dpi
    gate_offset = plot_params['gate_offset']

    renderer = obj.figure.canvas.get_renderer()
    width = obj.get_window_extent(renderer).width * factor + 2 * gate_offset
    height = obj.get_window_extent(renderer).height * factor + 2 * gate_offset

    axes.add_patch(
        Rectangle((gate_pos - width / 2, wire_pos - height / 2),
                  width,
                  height,
                  ec='k',
                  fc='w',
                  fill=True,
                  lw=plot_params['linewidth'],
                  zorder=6))


def draw_measure_gate(axes, gate_pos, wire_pos, plot_params):
    """
    Draws a measurement gate.

    Args:
        axes (AxesSubplot): axes object
        gate_pos (float): x coordinate of the gate [data units]
        wire_pos (float): y coordinate of the qubit wire
        plot_params (dict): plot parameters
    """
    # pylint: disable=invalid-name

    width = plot_params['mgate_width']
    height = 0.9 * width
    y_ref = wire_pos - 0.3 * height

    # Cannot use PatchCollection for the arc due to bug in matplotlib code...
    arc = Arc((gate_pos, y_ref),
              width * 0.7,
              height * 0.8,
              theta1=0,
              theta2=180,
              ec='k',
              fc='w',
              zorder=5)
    axes.add_patch(arc)

    patches = [
        Rectangle((gate_pos - width / 2, wire_pos - height / 2),
                  width,
                  height,
                  fill=True),
        Line2D((gate_pos, gate_pos + width * 0.35),
               (y_ref, wire_pos + height * 0.35),
               color='k',
               linewidth=1)
    ]

    gate = PatchCollection(patches,
                           edgecolors='k',
                           facecolors='w',
                           linewidths=plot_params['linewidth'],
                           zorder=5)
    gate.set_label('Measure')
    axes.add_collection(gate)


def multi_qubit_gate(axes, gate_str, gate_pos, wire_pos_min, wire_pos_max,
                     plot_params):
    """
    Draws a multi-target qubit gate.

    Args:
        axes (matplotlib.axes.Axes): axes object
        gate_str (str): string representation of a gate
        gate_pos (float): x coordinate of the gate [data units]
        wire_pos_min (float): y coordinate of the lowest qubit wire
        wire_pos_max (float): y coordinate of the highest qubit wire
        plot_params (dict): plot parameters
    """
    gate_offset = plot_params['gate_offset']
    y_center = (wire_pos_max - wire_pos_min) / 2 + wire_pos_min
    obj = axes.text(gate_pos,
                    y_center,
                    gate_str,
                    color='k',
                    ha='center',
                    va='center',
                    size=plot_params['fontsize'],
                    zorder=7)
    height = wire_pos_max - wire_pos_min + 2 * gate_offset
    inv = axes.transData.inverted()
    width = inv.transform_bbox(
        obj.get_window_extent(obj.figure.canvas.get_renderer())).width
    return axes.add_patch(
        Rectangle((gate_pos - width / 2, wire_pos_min - gate_offset),
                  width,
                  height,
                  edgecolor='k',
                  facecolor='w',
                  fill=True,
                  lw=plot_params['linewidth'],
                  zorder=6))


def draw_x_gate(axes, gate_pos, wire_pos, plot_params):
    """
    Draws the symbol for a X/NOT gate.

    Args:
        axes (matplotlib.axes.Axes): axes object
        gate_pos (float): x coordinate of the gate [data units]
        wire_pos (float): y coordinate of the qubit wire [data units]
        plot_params (dict): plot parameters
    """
    not_radius = plot_params['not_radius']

    gate = PatchCollection([
        Circle((gate_pos, wire_pos), not_radius, fill=False),
        Line2D((gate_pos, gate_pos),
               (wire_pos - not_radius, wire_pos + not_radius))
    ],
                           edgecolors='k',
                           facecolors='w',
                           linewidths=plot_params['linewidth'])
    gate.set_label('NOT')
    axes.add_collection(gate)


def draw_control_z_gate(axes, gate_pos, wire_pos1, wire_pos2, plot_params):
    """
    Draws the symbol for a controlled-Z gate.

    Args:
        axes (matplotlib.axes.Axes): axes object
        wire_pos (float): x coordinate of the gate [data units]
        y1 (float): y coordinate of the 1st qubit wire
        y2 (float): y coordinate of the 2nd qubit wire
        plot_params (dict): plot parameters
    """
    gate = PatchCollection([
        Circle(
            (gate_pos, wire_pos1), plot_params['control_radius'], fill=True),
        Circle(
            (gate_pos, wire_pos2), plot_params['control_radius'], fill=True),
        Line2D((gate_pos, gate_pos), (wire_pos1, wire_pos2))
    ],
                           edgecolors='k',
                           facecolors='k',
                           linewidths=plot_params['linewidth'])
    gate.set_label('CZ')
    axes.add_collection(gate)


def draw_swap_gate(axes, gate_pos, wire_pos1, wire_pos2, plot_params):
    """
    Draws the symbol for a SWAP gate.

    Args:
        axes (matplotlib.axes.Axes): axes object
        x (float): x coordinate [data units]
        y1 (float): y coordinate of the 1st qubit wire
        y2 (float): y coordinate of the 2nd qubit wire
        plot_params (dict): plot parameters
    """
    delta = plot_params['swap_delta']

    lines = []
    for wire_pos in (wire_pos1, wire_pos2):
        lines.append([(gate_pos - delta, wire_pos - delta),
                      (gate_pos + delta, wire_pos + delta)])
        lines.append([(gate_pos - delta, wire_pos + delta),
                      (gate_pos + delta, wire_pos - delta)])
    lines.append([(gate_pos, wire_pos1), (gate_pos, wire_pos2)])

    gate = LineCollection(lines,
                          colors='k',
                          linewidths=plot_params['linewidth'])
    gate.set_label('SWAP')
    axes.add_collection(gate)


def draw_wires(axes, n_labels, gate_grid, wire_grid, plot_params):
    """
    Draws all the circuit qubit wires.

    Args:
        axes (matplotlib.axes.Axes): axes object
        n_labels (int): number of qubit
        gate_grid (ndarray): array with the ref. x positions of the gates
        wire_grid (ndarray): array with the ref. y positions of the qubit
                             wires
        plot_params (dict): plot parameters
    """
    # pylint: disable=invalid-name

    lines = []
    for i in range(n_labels):
        lines.append(((gate_grid[0] - plot_params['column_spacing'],
                       wire_grid[i]), (gate_grid[-1], wire_grid[i])))
    all_lines = LineCollection(lines,
                               linewidths=plot_params['linewidth'],
                               edgecolor='k')
    all_lines.set_label('qubit_wires')
    axes.add_collection(all_lines)


def draw_labels(axes, qubit_labels, drawing_order, wire_grid, plot_params):
    """
    Draws the labels at the start of each qubit wire

    Args:
        axes (matplotlib.axes.Axes): axes object
        qubit_labels (list): labels of the qubit to be drawn
        drawing_order (dict): Mapping between wire indices and qubit IDs
        gate_grid (ndarray): array with the ref. x positions of the gates
        wire_grid (ndarray): array with the ref. y positions of the qubit
                             wires
        plot_params (dict): plot parameters
    """
    for qubit_id in qubit_labels:
        wire_idx = drawing_order[qubit_id]
        text(axes, plot_params['x_offset'], wire_grid[wire_idx],
             qubit_labels[qubit_id], plot_params)



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Defines a setup to compile to qubits placed in a linear chain or a circle.

It provides the `engine_list` for the `MainEngine`. This engine list contains
an AutoReplacer with most of the gate decompositions of ProjectQ, which are
used to decompose a circuit into only two qubit gates and arbitrary single
qubit gates. ProjectQ's LinearMapper is then used to introduce the necessary
Swap operations to route interacting qubits next to each other. This setup
allows to choose the final gate set (with some limitations).
"""

import inspect

import projectq
import projectq.libs.math
import projectq.setups.decompositions
from projectq.cengines import (AutoReplacer, DecompositionRuleSet,
                               InstructionFilter, LinearMapper, LocalOptimizer,
                               TagRemover)
from projectq.ops import (BasicMathGate, ClassicalInstructionGate, CNOT,
                          ControlledGate, get_inverse, QFT, Swap)


def high_level_gates(eng, cmd):
    """
    Remove any MathGates.
    """
    g = cmd.gate
    if g == QFT or get_inverse(g) == QFT or g == Swap:
        return True
    if isinstance(g, BasicMathGate):
        return False
    return True


def one_and_two_qubit_gates(eng, cmd):
    if isinstance(cmd.gate, ClassicalInstructionGate):
        # This is required to allow Measure, Allocate, Deallocate, Flush
        return True
    return len([q for qr in cmd.all_qubits for q in qr]) <= 2


def get_engine_list(num_qubits, cyclic=False, one_qubit_gates="any",
                    two_qubit_gates=(CNOT, Swap)):
    """
    Returns an engine list to compile to a linear chain of qubits.

    Note:
        If you choose a new gate set for which the compiler does not yet have
        standard rules, it raises an `NoGateDecompositionError` or a
        `RuntimeError: maximum recursion depth exceeded...`. Also note that
        even the gate sets which work might not yet be optimized. So make sure
        to double check and potentially extend the decomposition rules.
        This implemention currently requires that the one qubit gates must
        contain Rz and at least one of {Ry(best), Rx, H} and the two qubit gate
        must contain CNOT (recommended) or CZ.

    Note:
        Classical instructions gates such as e.g. Flush and Measure are
        automatically allowed.

    Example:
        get_engine_list(num_qubits=10, cyclic=False,
                        one_qubit_gates=(Rz, Ry, Rx, H),
                        two_qubit_gates=(CNOT,))

    Args:
        num_qubits(int): Number of qubits in the chain
        cyclic(bool): If a circle or not. Default is False
        one_qubit_gates: "any" allows any one qubit gate, otherwise provide
                         a tuple of the allowed gates. If the gates are
                         instances of a class (e.g. X), it allows all gates
                         which are equal to it. If the gate is a class (Rz), it
                         allows all instances of this class. Default is "any"
        two_qubit_gates: "any" allows any two qubit gate, otherwise provide
                         a tuple of the allowed gates. If the gates are
                         instances of a class (e.g. CNOT), it allows all gates
                         which are equal to it. If the gate is a class, it
                         allows all instances of this class.
                         Default is (CNOT, Swap).
    Raises:
        TypeError: If input is for the gates is not "any" or a tuple.

    Returns:
        A list of suitable compiler engines.
    """
    if two_qubit_gates != "any" and not isinstance(two_qubit_gates, tuple):
        raise TypeError("two_qubit_gates parameter must be 'any' or a tuple. "
                        "When supplying only one gate, make sure to correctly "
                        "create the tuple (don't miss the comma), "
                        "e.g. two_qubit_gates=(CNOT,)")
    if one_qubit_gates != "any" and not isinstance(one_qubit_gates, tuple):
        raise TypeError("one_qubit_gates parameter must be 'any' or a tuple.")

    rule_set = DecompositionRuleSet(modules=[projectq.libs.math,
                                             projectq.setups.decompositions])
    allowed_gate_classes = []
    allowed_gate_instances = []
    if one_qubit_gates != "any":
        for gate in one_qubit_gates:
            if inspect.isclass(gate):
                allowed_gate_classes.append(gate)
            else:
                allowed_gate_instances.append((gate, 0))
    if two_qubit_gates != "any":
        for gate in two_qubit_gates:
            if inspect.isclass(gate):
                #  Controlled gate classes don't yet exists and would require
                #  separate treatment
                assert not isinstance(gate, ControlledGate)
                allowed_gate_classes.append(gate)
            else:
                if isinstance(gate, ControlledGate):
                    allowed_gate_instances.append((gate._gate, gate._n))
                else:
                    allowed_gate_instances.append((gate, 0))
    allowed_gate_classes = tuple(allowed_gate_classes)
    allowed_gate_instances = tuple(allowed_gate_instances)

    def low_level_gates(eng, cmd):
        all_qubits = [q for qr in cmd.all_qubits for q in qr]
        assert len(all_qubits) <= 2
        if isinstance(cmd.gate, ClassicalInstructionGate):
            # This is required to allow Measure, Allocate, Deallocate, Flush
            return True
        if one_qubit_gates == "any" and len(all_qubits) == 1:
            return True
        if two_qubit_gates == "any" and len(all_qubits) == 2:
            return True
        if isinstance(cmd.gate, allowed_gate_classes):
            return True
        if (cmd.gate, len(cmd.control_qubits)) in allowed_gate_instances:
            return True
        return False

    return [AutoReplacer(rule_set),
            TagRemover(),
            InstructionFilter(high_level_gates),
            LocalOptimizer(5),
            AutoReplacer(rule_set),
            TagRemover(),
            InstructionFilter(one_and_two_qubit_gates),
            LocalOptimizer(5),
            LinearMapper(num_qubits=num_qubits, cyclic=cyclic),
            AutoReplacer(rule_set),
            TagRemover(),
            InstructionFilter(low_level_gates),
            LocalOptimizer(5),
            ]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains a TagRemover engine, which removes temporary command tags (such as
Compute/Uncompute), thus enabling optimization across meta statements (loops
after unrolling, compute/uncompute, ...)
"""
from projectq.cengines import BasicEngine
from projectq.meta import ComputeTag, UncomputeTag


class TagRemover(BasicEngine):
    """
    TagRemover is a compiler engine which removes temporary command tags (see
    the tag classes such as LoopTag in projectq.meta._loop).

    Removing tags is important (after having handled them if necessary) in
    order to enable optimizations across meta-function boundaries (compute/
    action/uncompute or loops after unrolling)
    """
    def __init__(self, tags=[ComputeTag, UncomputeTag]):
        """
        Construct the TagRemover.

        Args:
            tags: A list of meta tag classes (e.g., [ComputeTag, UncomputeTag])
                denoting the tags to remove
        """
        BasicEngine.__init__(self)
        assert isinstance(tags, list)
        self._tags = tags

    def receive(self, command_list):
        """
        Receive a list of commands from the previous engine, remove all tags
        which are an instance of at least one of the meta tags provided in the
        constructor, and then send them on to the next compiler engine.

        Args:
            command_list (list<Command>): List of commands to receive and then
                (after removing tags) send on.
        """
        for cmd in command_list:
            for tag in self._tags:
                cmd.tags = [t for t in cmd.tags if not isinstance(t, tag)]
            self.send([cmd])



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition rule for the phase-shift gate.

Decomposes the (controlled) phase-shift gate using z-rotation and a global
phase gate.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import Control
from projectq.ops import Ph, Rz, R


def _decompose_R(cmd):
    """ Decompose the (controlled) phase-shift gate, denoted by R(phase). """
    ctrl = cmd.control_qubits
    eng = cmd.engine
    gate = cmd.gate

    with Control(eng, ctrl):
        Ph(.5 * gate.angle) | cmd.qubits
        Rz(gate.angle) | cmd.qubits


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(R, _decompose_R)
]
